'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
import tokenize
import token
import StringIO

class TokenizerError(Exception):
    """Base class for All exceptions"""

    def __init__(self, a_msg, a_line=None, a_col=None):
        
        self._line = a_line
        self._col  = a_col
        
        if self._line == None and self._col == None:
            extra = "" 
        else:
            extra = "(line=%s,col=%s)" % (self._line, self._col)
        
        super(TokenizerError, self).__init__("%s %s." % (a_msg, extra))
    

class Token(object):
    """ Token class """
    def __init__(self, a_type, num, value, begin, end, parsed_line):
        
        self._type  = a_type
        self._num   = num
        self._value = value
        self._begin = begin
        self._end   = end
        self._parsed_line  = parsed_line
    
    @property
    def type(self):
        """ Return the token type """
        return self._type

    @property
    def num(self):
        """ Return the token type num """
        return self._num

    @property
    def value(self):
        """ Return the token value """
        return self._value
    
    @property
    def begin(self):
        """ Return the token begin """
        return self._begin
    
    @property
    def end(self):
        """ Return the token end """
        return self._end
    
    @property
    def parsed_line(self):
        """ Return the token line """
        return self._parsed_line
    
    def __repr__(self):
        return "[type,num]=[%s,%s],value=[%s], parsed line=%s,[begin index,end index]=[%s,%s]" \
               % (self._type, self._num, self._value, self._parsed_line, self._begin, self._end)
         

class Tokenizer(object):
    """ 
        Create tokens for parsing the grammar. 
        This class is a wrapper around the python tokenizer adapt to the DSL that is going to be used.
    """    
    def __init__(self):
        """ constructor """
        # list of tokens
        self._tokens  = []
        
        self._index   = 0
        
        self._current = None
       
    def tokenize(self, a_program, a_eatable_token_types = ()):
        """ parse the expression.
            By default the parser eats space but some extra tokens by have to be eaten
        
            Args:
               a_expression: the expression to parser
               
            Returns:
               return dict containing the different parts of the request (spectrum, ....)
        
            Raises:
               exception TokenizerError if the syntax of the aString string is incorrect
        """
        g_info = tokenize.generate_tokens(StringIO.StringIO(a_program).readline)   # tokenize the string
        
        for toknum, tokval, tokbeg, tokend, tokline  in g_info:
            if token.tok_name[toknum] not in a_eatable_token_types:
                self._tokens.append(Token(token.tok_name[toknum], toknum, tokval, tokbeg, tokend, tokline))
            
        
            
    def __iter__(self):
        """ iterator implemented with a generator.
        """
        for tok in self._tokens:
            self._current = tok
            yield tok
        
    def next(self):
        """ get next token.
          
            Returns:
               return next token
        """
        
        self._current = self._tokens[self._index]
        self._index += 1
        return self._current
    
    def has_next(self):
        """ check it there are more tokens to consume.
        
            Returns:
               return True if more tokens to consume False otherwise
        """
        return self._index < len(self._tokens)
    
    def current_token(self):
        """ return the latest consumed token.
        
            Returns:
               return the latest consumerd token
        """
        return self._current
    
    def consume_token(self, what):
        """ consume the next token if it is what """
        if self._current.value != what :
            raise TokenizerError("Expected '%s' but instead found '%s'" % (what, self._current.value))
        else:
            return self.next()
        
    def consume_while_next_token_is_in(self, a_token_types_list):
        """
           Consume the next tokens as long as they have one of the passed types.
           This means that at least one token with one of the passed types needs to be matched.
           
           Args:
               a_token_types_list: the token types to consume
            
           Returns:
               return the next non matching token 
        """
        
        self.consume_next_tokens(a_token_types_list)
        
        while True:
        
            tok = self.next()
        
            if tok.type not in a_token_types_list:
                return tok
    
    def consume_while_current_token_is_in(self, a_token_types_list): #pylint: disable=C0103
        """
           Consume the tokens starting from the current token as long as they have one of the passed types.
           It is a classical token eater. It eats tokens as long as they are the specified type
           
           Args:
               a_token_types_list: the token types to consume
            
           Returns:
               return the next non matching token 
        """
        
        tok = self.current_token()
        
        while tok.type in a_token_types_list:
            tok = self.next()
        
        return tok
    
    def consume_next_tokens(self, a_token_types_list):
        """
           Consume one of the next token types given in the list and check that it is the expected type otherwise send an exception
            
           Args:
               a_tokens_list:  the token types to list 
               
           Returns:
               return next token 
           
           Raises:
               exception  BadTokenError if a Token Type that is not in a_token_types_list is found
        """
        
        tok = self.next()
        
        if tok.type not in a_token_types_list:
            raise TokenizerError("Expected '%s' but instead found '%s'" % (a_token_types_list, tok))
        else:
            return tok
    
    def advance(self, inc=1):
        """ return the next + inc token but do not consume it.
            Useful to check future tokens.
        
            Args:
               a_expression: increment + 1 is the default (just look one step forward)
               
            Returns:
               return lookhead token
        """
        return self._tokens[self._index-1 + inc]
    
class CompilerError(Exception):
    """Base class for All exceptions"""

    def __init__(self, a_msg, a_line=None, a_col=None):
        
        self._line = a_line
        self._col  = a_col
        
        msg = ''
        
        if self._line == None and self._col == None:
            extra = ""
            msg = "%s." % (a_msg) 
        else:
            extra = "(line=%s,col=%s)" % (self._line, self._col)
            msg = "%s %s." % (a_msg, extra)
        
        super(CompilerError, self).__init__(msg)
    
class Compiler(object):
    """ compile some python structures
    """
    
    def __init__(self):
        """ constructor """
        
        #default tokens to ignore
        self._tokens_to_ignore = ('INDENT', 'DEDENT', 'NEWLINE', 'NL')
    
    def compile_list(self, a_to_compile_str):
        """ compile a list object """
        
        try:
            tokenizer = Tokenizer()
            tokenizer.tokenize(a_to_compile_str, self._tokens_to_ignore)
        except tokenize.TokenError, err:
            
            #translate this error into something understandable. 
            #It is because the bloody tokenizer counts the brackets
            if err.args[0] == "EOF in multi-line statement":
                raise CompilerError("Expression \"%s\" cannot be converted as a list" % (a_to_compile_str))
            else:
                raise CompilerError(err)
            
            print("Err = %s\n" % (err))
        
        tokenizer.next()
        
        return self._compile_list(tokenizer)
    
    def compile_dict(self, a_to_compile_str):
        """ compile a dict object """
        
        try:
            tokenizer = Tokenizer()
            tokenizer.tokenize(a_to_compile_str, self._tokens_to_ignore)
        except tokenize.TokenError, err:
            
            #translate this error into something understandable. 
            #It is because the bloody tokenizer counts the brackets
            if err.args[0] == "EOF in multi-line statement":
                raise CompilerError("Expression \"%s\" cannot be converted as a dict" % (a_to_compile_str))
            else:
                raise CompilerError(err)
            
            print("Err = %s\n" % (err))
        
        tokenizer.next()
        
        return self._compile_dict(tokenizer)

    def _compile_dict(self, a_tokenizer):
        """ internal method for compiling a dict struct """
        result = {}
        
        the_token = a_tokenizer.current_token()
        
        while the_token.type != 'ENDMARKER':
            
            #look for an open bracket
            if the_token.type == 'OP' and the_token.value == '{':
               
                the_token = a_tokenizer.next()
                
                while True:
                   
                    if the_token.type == 'OP' and the_token.value == '}':
                        return result
                    else:
                        # get key values
                        (key, val) = self._compile_key_value(a_tokenizer)

                        result[key] = val  
                    
                    the_token = a_tokenizer.current_token()
                                   
            else:
                raise CompilerError("Unsupported token (type: %s, value : %s)" \
                                    % (the_token.type, the_token.value), the_token.begin[0], the_token.begin[1])
            
        #we should never reach that point (compilation error)
        raise CompilerError("End of line reached without finding a list. The line [%s] cannot be transformed as a list" \
                            % (the_token.parsed_line))
        
    def _compile_key_value(self, a_tokenizer):
        """ look for the pair key value component of a dict """
        
        the_token = a_tokenizer.current_token()
        
        key = None
        val = None
        
        # get key
        if the_token.type in ('STRING', 'NUMBER', 'NAME'):
            
            #next the_token is in _compile_litteral
            key = self._compile_litteral(a_tokenizer)
            
            the_token = a_tokenizer.current_token()
            
        else:
            raise CompilerError("unexpected token (type: %s, value : %s)" \
                                % (the_token.type, the_token.value), \
                                the_token.begin[0], the_token.begin[1])  
        
        #should have a comma now
        if the_token.type != 'OP' and the_token.value != ':':
            raise CompilerError("Expected a token (type:OP, value: :) but instead got (type: %s, value: %s)" \
                                % (the_token.type, the_token.value), the_token.begin[0], the_token.begin[1])
        else:
            #eat it
            the_token = a_tokenizer.next()
        
        #get value
        # it can be a
        if the_token.type in ('STRING', 'NUMBER', 'NAME'):
            #next the_token is in _compile_litteral
            val = self._compile_litteral(a_tokenizer)
            
            the_token = a_tokenizer.current_token()
        
        #check for a list
        elif the_token.value == '[' and the_token.type == 'OP':
            
            # look for a list
            val = self._compile_list(a_tokenizer)
            
            # positioning to the next token
            the_token = a_tokenizer.next()
            
        elif the_token.value == '{' and the_token.type == 'OP':
            
            # look for a dict
            val = self._compile_dict(a_tokenizer)
            
            # positioning to the next token
            the_token = a_tokenizer.next()
        
        elif the_token.value == '(' and the_token.type == 'OP':
            
            # look for a dict
            val = self._compile_tuple(a_tokenizer)
            
            # positioning to the next token
            the_token = a_tokenizer.next()
            
        else:
            raise CompilerError("unexpected token (type: %s, value : %s)" \
                                % (the_token.type, the_token.value), the_token.begin[0], \
                                the_token.begin[1])  
        
        #if we have a comma then eat it as it means that we will have more than one values
        if the_token.type == 'OP' and the_token.value == ',':
            the_token = a_tokenizer.next() 
            
        return (key, val)               
        
        
    def _compile_litteral(self, a_tokenizer):
        """ compile key. A key can be a NAME, STRING or NUMBER """
        
        val   = None
        
        dummy = None
        
        the_token = a_tokenizer.current_token()
        
        while the_token.type not in ('OP', 'ENDMARKER'):
            if the_token.type == 'STRING':  
                #check if the string is unicode
                if len(the_token.value) >= 3 and the_token.value[:2] == "u'":
                    #unicode string
                    #dummy = unicode(the_token.value[2:-1], 'utf_8') #decode from utf-8 encoding not necessary if read full utf-8 file
                    dummy = unicode(the_token.value[2:-1])
                else:
                    #ascii string
                    # the value contains the quote or double quotes so remove them always
                    dummy = the_token.value[1:-1]
                    
            elif the_token.type == 'NAME':
                # intepret all non quoted names as a string
                dummy = the_token.value
                    
            elif the_token.type == 'NUMBER':  
                     
                dummy = self._create_number(the_token.value)
                 
            else:
                raise CompilerError("unexpected token (type: %s, value : %s)" \
                                    % (the_token.type, the_token.value), \
                                    the_token.begin[0], the_token.begin[1])
           
            #if val is not None, it has to be a string
            if val:
                val = '%s %s' % (str(val), str(dummy))
            else:
                val = dummy
            
            the_token = a_tokenizer.next()
            
        return val
    
    
    def _compile_tuple(self, a_tokenizer):
        """ process tuple structure """
        result = []
        
        open_bracket = 0
        # this is the mode without [ & ] operator : 1,2,3,4
        simple_list_mode = 0
        
        the_token = a_tokenizer.current_token()
        
        while the_token.type != 'ENDMARKER':
            #look for an open bracket
            if the_token.value == '(' and the_token.type == 'OP':
                #first time we open a bracket and not in simple mode 
                if open_bracket == 0 and simple_list_mode == 0:
                    open_bracket += 1
                #recurse to create the imbricated list
                else:
                    result.append(self._compile_tuple(a_tokenizer))
                    
                the_token = a_tokenizer.next()
            
            elif the_token.value == '{' and the_token.type == 'OP':
               
                result.append(self._compile_dict(a_tokenizer))
                    
                the_token = a_tokenizer.next()
            
            elif the_token.value == '[' and the_token.type == 'OP':
               
                result.append(self._compile_list(a_tokenizer))
                    
                the_token = a_tokenizer.next()
                    
            elif the_token.type == 'OP' and the_token.value == ')':
                # end of list return result
                if open_bracket == 1:
                    return tuple(result)
                # cannot find a closing bracket and a simple list mode
                elif simple_list_mode == 1:
                    raise CompilerError("unexpected token (type: %s, value : %s)" \
                                        % (the_token.value, the_token.type), the_token.begin[0], \
                                        the_token.begin[1])
            # the comma case
            elif the_token.type == 'OP' and the_token.value == ',':
                # just eat it
                the_token = a_tokenizer.next()
                
            elif the_token.type in ('STRING', 'NUMBER', 'NAME'):
                
                # find values outside of a list 
                # this can be okay
                if open_bracket == 0:
                    simple_list_mode = 1
                    
                #next the_token is in _compile_litteral
                result.append(self._compile_litteral(a_tokenizer))
                
                the_token = a_tokenizer.current_token()
               
            else:
                raise CompilerError("Unsupported token (type: %s, value : %s)"\
                                    % (the_token.value, the_token.type), \
                                    the_token.begin[0], the_token.begin[1])
            
        
        # if we are in simple_list_mode return list else error
        if simple_list_mode == 1:
            return tuple(result)
            
        #we should never reach that point (compilation error)
        raise CompilerError("End of line reached without finding a list. The line [%s] cannot be transformed as a tuple" \
                            % (the_token.parsed_line))
    
    def _compile_list(self, a_tokenizer):
        """ process a list structure """
        result = []
        
        
        open_bracket = 0
        # this is the mode without [ & ] operator : 1,2,3,4
        simple_list_mode = 0
        
        the_token = a_tokenizer.current_token()
        
        while the_token.type != 'ENDMARKER':
            #look for an open bracket
            if the_token.value == '[' and the_token.type == 'OP':
                #first time we open a bracket and not in simple mode 
                if open_bracket == 0 and simple_list_mode == 0:
                    open_bracket += 1
                #recurse to create the imbricated list
                else:
                    result.append(self._compile_list(a_tokenizer))
                    
                the_token = a_tokenizer.next()
            
            elif the_token.value == '(' and the_token.type == 'OP':
               
                result.append(self._compile_tuple(a_tokenizer))
                    
                the_token = a_tokenizer.next()
            
            elif the_token.value == '{' and the_token.type == 'OP':
               
                result.append(self._compile_dict(a_tokenizer))
                    
                the_token = a_tokenizer.next()
                    
            elif the_token.type == 'OP' and the_token.value == ']':
                # end of list return result
                if open_bracket == 1:
                    return result
                # cannot find a closing bracket and a simple list mode
                elif simple_list_mode == 1:
                    raise CompilerError("unexpected token (type: %s, value : %s)" \
                                        % (the_token.value, the_token.type), the_token.begin[0], the_token.begin[1])
            # the comma case
            elif the_token.type == 'OP' and the_token.value == ',':
                # just eat it
                the_token = a_tokenizer.next()
                
            elif the_token.type in ('STRING', 'NUMBER', 'NAME'):
                
                # find values outside of a list 
                # this can be okay
                if open_bracket == 0:
                    simple_list_mode = 1
                    
                #next the_token is in _compile_litteral
                result.append(self._compile_litteral(a_tokenizer))
                
                the_token = a_tokenizer.current_token()
               
            else:
                raise CompilerError("Unsupported token (type: %s, value : %s)"\
                                    % (the_token.value, the_token.type), \
                                    the_token.begin[0], the_token.begin[1])
            
        
        # if we are in simple_list_mode return list else error
        if simple_list_mode == 1:
            return result
            
        #we should never reach that point (compilation error)
        raise CompilerError("End of line reached without finding a list. The line [%s] cannot be transformed as a list" \
                            % (the_token.parsed_line))
         
    @classmethod
    def _create_number(cls, a_number):
        """ depending on the value return a int or a float. 
            For the moment very simple: If there is . it is a float"""
        
        if a_number.find('.') > 0:
            return float(a_number)
        else:
            return int(a_number)



# -*- coding: utf-8 -*-
'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''

# unit tests part
import unittest
import sys
import os
import codecs
import gmv.conf.conf_helper


class TestConf(unittest.TestCase): #pylint: disable=R0904
    """
       Test Class for the Conf Object
    """  
    @classmethod
    def _get_tests_dir_path(cls):
        """ get the org.ctbto.conf.tests path depending on where it is defined """
        
        fmod_path = gmv.conf.__path__
        
        test_dir = "%s/tests" % fmod_path[0]
        
        return test_dir
    
    def setUp(self): #pylint: disable=C0103
         
        # necessary for the include with the VAR ENV substitution
        os.environ["DIRCONFENV"] = TestConf._get_tests_dir_path()
         
        self.conf = gmv.conf.conf_helper.Conf(use_resource=False)
    
        #the_fp = open('%s/%s' % (TestConf._get_tests_dir_path(), "test.config"))
        the_fp = codecs.open('%s/%s' % (TestConf._get_tests_dir_path(), "test.config"), 'r', 'utf-8')
    
        self.conf._read(the_fp,"the file") #pylint: disable=W0212

    def tearDown(self): #pylint: disable=C0103

        if os.path.exists('/tmp/fake_conf.config'):
            os.remove('/tmp/fake_conf.config')
 
    def test_empty(self):
        """
          Do nothing
        """
        pass
        
    def test_get_objects(self):
        """testGetObjects: test getter from all types """
        # get simple string
        astring = self.conf.get("GroupTest1", "astring")
        
        self.assertEqual(astring,"oracle.jdbc.driver.OracleDriver")
        
        # get an int
        aint = self.conf.getint("GroupTest1", "aint")
        
        self.assertEqual(aint, 10)
        
        # get floatcompile the statements
        afloat = self.conf.getfloat("GroupTest1", "afloat")
        
        self.assertEqual(afloat, 5.24)
        
        # get different booleans form
        abool1 = self.conf.getboolean("GroupTest1", "abool1")
        
        self.assertEqual(abool1, True)
        
        abool2 = self.conf.getboolean("GroupTest1", "abool2")
        
        self.assertEqual(abool2, False)
        
        abool3 = self.conf.getboolean("GroupTest1", "abool3")
        
        self.assertEqual(abool3, True)
        
        abool4 = self.conf.getboolean("GroupTest1", "abool4")
        
        self.assertEqual(abool4 , False)
        
    def test_get_defaults(self):
        """testGetDefaults: test defaults values """
        
        # get all defaults
        astring = self.conf.get("GroupTest", "astring", "astring")
        
        self.assertEqual(astring, "astring")
        
        # get an default for int
        aint = self.conf.getint("GroupTest", "aint", 2)
        
        self.assertEqual(aint, 2)
        
        # get float
        afloat = self.conf.getfloat("GroupTest", "afloat", 10.541)
        
        self.assertEqual(afloat, 10.541)
        
        abool1 = self.conf.getboolean("GroupTest", "abool1", True)
        
        self.assertEqual(abool1, True)
        
        abool2 = self.conf.getboolean("GroupTest", "abool2", False)
        
        self.assertEqual(abool2, False)
        
        # existing group no option
        abool5 = self.conf.getboolean("GroupTest1", "abool32", False)
        
        self.assertEqual(abool5, False)
        
    def test_var_substitutions(self):
        """testVarSubstitutions: test variables substitutions"""
        
        # simple substitution
        apath = self.conf.get("GroupTestVars", "path")
        
        self.assertEqual(apath,"/foo/bar//tmp/foo/bar/bar/foo")
        
        # multiple substitution
        apath = self.conf.get("GroupTestVars", "path1")
        
        self.assertEqual(apath,"/foo//tmp/foo/bar//foo/bar//tmp/foo/bar/bar/foo/bar")
        
        # nested substitution
        nested = self.conf.get("GroupTestVars", "nested")
        
        self.assertEqual(nested, "this is done")  
        
    def test_include(self):
        """testInclude: test includes """
        val = self.conf.get("IncludedGroup", "hello")
        
        self.assertEqual(val, 'foo')
        
    @classmethod
    def _create_fake_conf_file_in_tmp(cls):
        """Create a fake conf file in tmp"""
        the_f = open('/tmp/fake_conf.config', 'w')
        
        the_f.write('\n[MainDatabaseAccess]\n')
        the_f.write('driverClassName=oracle.jdbc.driver.OracleDriver')
        the_f.flush()
        the_f.close()
    
    def ztest_use_conf_ENVNAME_resource(self): #pylint: disable=C0103
        """testUseConfENVNAMEResource: Use default resource ENVNAME to locate conf file"""
        self._create_fake_conf_file_in_tmp()

        # need to setup the ENV containing the the path to the conf file:
        os.environ[gmv.conf.conf_helper.Conf.ENVNAME] = "/tmp/fake_conf.config"
   
        self.conf = gmv.conf.conf_helper.Conf.get_instance()
        
        the_s = self.conf.get("MainDatabaseAccess", "driverClassName")
        
        self.assertEqual(the_s, 'oracle.jdbc.driver.OracleDriver')
    
    def test_read_from_CLI(self): #pylint: disable=C0103
        """testReadFromCLI: do substitutions from command line resources"""
        #set environment
        os.environ["TESTENV"] = "/tmp/foo/foo.bar"
        
        val = self.conf.get("GroupTest1", "fromenv")
   
        self.assertEqual(val, '/mydir//tmp/foo/foo.bar')
        
        #set cli arg
        sys.argv.append("--LongName")
        sys.argv.append("My Cli Value")
        
        val = self.conf.get("GroupTest1", "fromcli1")
   
        self.assertEqual(val, 'My Cli Value is embedded')
        
        #check with a more natural cli value
        val = self.conf.get("GroupTest1", "fromcli2")
   
        self.assertEqual(val, 'My Cli Value is embedded 2')
    
    def test_read_from_ENV(self): #pylint: disable=C0103
        """testReadFromENV: do substitutions from ENV resources"""
        #set environment
        os.environ["TESTENV"] = "/tmp/foo/foo.bar"
        
        val = self.conf.get("ENV", "TESTENV")
        
        self.assertEqual(val, "/tmp/foo/foo.bar")
        
        #set cli arg
        sys.argv.append("--LongName")
        sys.argv.append("My Cli Value")
        
        val = self.conf.get("CLI", "LongName")
        
        self.assertEqual(val, "My Cli Value")
        
        # get a float from env
        os.environ["TESTENV"] = "1.05"
        
        val = self.conf.getfloat("ENV", "TESTENV")
        
        self.assertEqual(val+1, 2.05) 
    
    def test_print_content(self):
        """ test print content """
        
        #set environment
        os.environ["TESTENV"] = "/tmp/foo/foo.bar"
        
        #set cli arg
        sys.argv.append("--LongName")
        sys.argv.append("My Cli Value")
        
        substitute_values = True
        
        result = self.conf.print_content( substitute_values )
        
        self.assertNotEqual(result, '')
        
    def test_value_as_List(self): #pylint: disable=C0103
        """ Value as List """
        
        the_list = self.conf.getlist('GroupTestValueStruct', 'list')
        
        self.assertEqual(the_list, ['a', 1, 3])
    
    def test_value_as_unicodeList(self): #pylint: disable=C0103
        """ Value as List """
        
        the_list = self.conf.getlist('GroupTestValueStruct', 'unicode_list')
        
        self.assertEqual(the_list, [ u'[Gmail]/Чаты', 'z' , 1 ])
    
    def test_value_as_dict(self):
        """Dict as Value """
        
        the_dict = self.conf.get_dict('GroupTestValueStruct', 'dict')
        
        self.assertEqual(the_dict, {'a': 2, 'b': 3})
    
    def test_complex_dict(self):
        """ complex dict """
        the_dict = self.conf.get_dict('GroupTestValueStruct', 'complex_dict')
        
        self.assertEqual(the_dict, {'a': 2, 'c': {'a': 1, 'c': [1, 2, 3], 'b': [1, 2, 3, 4, 5, 6, 7]}, 'b': 3})
    
    def test_dict_error(self):
        """ error with a dict """
        
        try:
            self.conf.get_dict('GroupTestValueStruct', 'dict_error')
        except Exception, err:
            self.assertEquals(err.message, "Expression \"{1:2,'v b': a\" cannot be converted as a dict.")
            return
        
        self.fail('Should never reach that point')
            
    def test_list_error(self):
        """ error with a list """
        
        try:
            the_list = self.conf.get_list('GroupTestValueStruct', 'list_error')
            print('the_list = %s\n' % (the_list))
        except Exception, err:
            self.assertEquals(err.message, 'Unsupported token (type: @, value : OP) (line=1,col=3).')
            return
         
        self.fail('Should never reach that point')
        
class TestResource(unittest.TestCase):
    """
       Test Class for the Resource object
    """   
    def test_resource_simple_cli(self):
        """testResourceSimpleCli: read resource from CLI"""
        # set command line
        sys.argv.append("--LongName")
        sys.argv.append("My Cli Value")
        
        resource = gmv.conf.conf_helper.Resource(a_cli_argument = "--LongName", a_env_variable = None) 
        
        self.assertEqual("My Cli Value", resource.get_value())
        
        # look for LongName without --. It should be formalized by the Resource object
        resource = gmv.conf.conf_helper.Resource(a_cli_argument = "LongName", a_env_variable = None) 
        
        self.assertEqual("My Cli Value", resource.get_value())
    
    def test_resource_from_env(self): 
        """testResourceFromENV: read resource from ENV"""   
        #ENV 
        os.environ["MYENVVAR"] = "My ENV Value"
  
        resource = gmv.conf.conf_helper.Resource(a_cli_argument=None, a_env_variable="MYENVVAR")
        
        self.assertEqual("My ENV Value", resource.get_value())
        
    def ztest_resource_priority_rules(self):
        """testResourcePriorityRules: test priority rules"""   
        resource = gmv.conf.conf_helper.Resource(a_cli_argument="--LongName", a_env_variable="MYENVVAR")
  
        self.assertEqual("My Cli Value", resource.get_value())
  
    def test_resource_get_different_types(self): #pylint: disable=C0103
        """testResourceGetDifferentTypes: return resource in different types"""
        
        os.environ["MYENVVAR"] = "yes"
        resource = gmv.conf.conf_helper.Resource(a_cli_argument=None, a_env_variable="MYENVVAR")
        
        self.assertEqual(resource.get_value_as_boolean(), True)
        
        os.environ["MYENVVAR"] = "4"
  
        resource = gmv.conf.conf_helper.Resource(a_cli_argument=None, a_env_variable="MYENVVAR")
  
        self.assertEqual(resource.get_value_as_int()+1, 5)
        
        os.environ["MYENVVAR"] = "4.345"
  
        resource = gmv.conf.conf_helper.Resource(a_cli_argument=None, a_env_variable="MYENVVAR")
  
        self.assertEqual(resource.get_value_as_float()+1, 5.345)
        
def tests():
    """ global test method"""
    #suite = unittest.TestLoader().loadTestsFromModule(gmv.conf.conf_tests)
    suite = unittest.TestLoader().loadTestsFromTestCase(TestConf)
    unittest.TextTestRunner(verbosity=2).run(suite)
 
        
if __name__ == '__main__':
    tests()



'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 Module handling the xauth authentication.
 Strongly influenced by http://code.google.com/p/googlecl/source/browse/trunk/src/googlecl/service.py
 and xauth part of gyb http://code.google.com/p/got-your-back/source/browse/trunk/gyb.py

'''
import gdata.service
import webbrowser
import random
import time
import atom
import urllib

import os
import getpass

import gmv.log_utils as log_utils
import gmv.blowfish as blowfish
import gmv.gmvault_utils as gmvault_utils

LOG = log_utils.LoggerFactory.get_logger('oauth')


def get_2_legged_oauth_tok_sec():
    '''
       Get 2 legged token and secret
    '''
    tok = raw_input('Enter your domain\'s OAuth consumer key: ')
  
    sec = raw_input('Enter your domain\'s OAuth consumer secret: ')
      
    return tok, sec, "two_legged"
    

def get_oauth_tok_sec(email, use_webbrowser = False, debug=False):
    '''
       Generate token and secret
    '''
    
    scopes = ['https://mail.google.com/', # IMAP/SMTP client access
              'https://www.googleapis.com/auth/userinfo#email'] # Email address access (verify token authorized by correct account
    
    gdata_serv = gdata.service.GDataService()
    gdata_serv.debug = debug
    gdata_serv.source = 'gmvault '
    
    gdata_serv.SetOAuthInputParameters(gdata.auth.OAuthSignatureMethod.HMAC_SHA1, \
                                       consumer_key = 'anonymous', consumer_secret = 'anonymous')
    
    params = {'xoauth_displayname':'Gmvault - Backup your Gmail account'}
    try:
        request_token = gdata_serv.FetchOAuthRequestToken(scopes=scopes, extra_parameters = params)
    except gdata.service.FetchingOAuthRequestTokenFailed, err:
        if str(err).find('Timestamp') != -1:
            LOG.critical('Is your system clock up to date? See the FAQ http://code.google.com/p/googlecl/wiki/FAQ'\
                         '#Timestamp_too_far_from_current_time\n')
            
        LOG.critical("Received Error: %s.\n" % (err) )
        LOG.critical("=== Exception traceback ===")
        LOG.critical(gmvault_utils.get_exception_traceback())
        LOG.critical("=== End of Exception traceback ===\n")
            
        return (None, None)
    
    url_params = {}
    domain = email[email.find('@')+1:]
    if domain.lower() != 'gmail.com' and domain.lower() != 'googlemail.com':
        url_params = {'hd': domain}
    
    auth_url = gdata_serv.GenerateOAuthAuthorizationURL(request_token=request_token, extra_params=url_params)
    
    #message to indicate that a browser will be opened
    raw_input('gmvault will now open a web browser page in order for you to grant gmvault access to your Gmail.\n'\
              'Please make sure you\'re logged into the correct Gmail account (%s) before granting access.\n'\
              'Press ENTER to open the browser. Once you\'ve granted access you can switch back to gmvault.' % (email))
    
    # run web browser otherwise print message with url
    if use_webbrowser:
        try:
            webbrowser.open(str(auth_url))  
        except Exception, err: #pylint: disable-msg=W0703
            LOG.critical("Error: %s.\n" % (err) )
            LOG.critical("=== Exception traceback ===")
            LOG.critical(gmvault_utils.get_exception_traceback())
            LOG.critical("=== End of Exception traceback ===\n")
        
        raw_input("You should now see the web page on your browser now.\n"\
                  "If you don\'t, you can manually open:\n\n%s\n\nOnce you've granted"\
                  " gmvault access, press the Enter key.\n" % (auth_url))
        
    else:
        raw_input('Please log in and/or grant access via your browser at %s '
                  'then hit enter.' % (auth_url))
    
    try:
        final_token = gdata_serv.UpgradeToOAuthAccessToken(request_token)
    except gdata.service.TokenUpgradeFailed:
        LOG.critical('Token upgrade failed! Could not get OAuth access token.\n Did you grant gmvault access in your browser ?')
        LOG.critical("=== Exception traceback ===")
        LOG.critical(gmvault_utils.get_exception_traceback())
        LOG.critical("=== End of Exception traceback ===\n")
        
        return (None, None)

    return (final_token.key, final_token.secret, "normal")

def generate_xoauth_req(a_token, a_secret, email, type):
    """
       generate the xoauth req from a user token and secret.
       Handle two_legged xoauth for admins.
    """
    nonce = str(random.randrange(2**64 - 1))
    timestamp = str(int(time.time()))
    if type == "two_legged": #2 legged oauth
        request = atom.http_core.HttpRequest('https://mail.google.com/mail/b/%s/imap/?xoauth_requestor_id=%s' \
                                             % (email, urllib.quote(email)), 'GET')
         
        signature = gdata.gauth.generate_hmac_signature(http_request=request, consumer_key=a_token, consumer_secret=a_secret, \
                                                        timestamp=timestamp, nonce=nonce, version='1.0', next=None)
        return '''GET https://mail.google.com/mail/b/%s/imap/?xoauth_requestor_id=%s oauth_consumer_key="%s",oauth_nonce="%s"'''\
               ''',oauth_signature="%s",oauth_signature_method="HMAC-SHA1",oauth_timestamp="%s",oauth_version="1.0"''' \
               % (email, urllib.quote(email), a_token, nonce, urllib.quote(signature), timestamp)
    else:
        request = atom.http_core.HttpRequest('https://mail.google.com/mail/b/%s/imap/' % email, 'GET')
        signature = gdata.gauth.generate_hmac_signature(
            http_request=request, consumer_key='anonymous', consumer_secret='anonymous', timestamp=timestamp,
            nonce=nonce, version='1.0', next=None, token = a_token, token_secret= a_secret)
        return '''GET https://mail.google.com/mail/b/%s/imap/ oauth_consumer_key="anonymous",oauth_nonce="%s"'''\
               ''',oauth_signature="%s",oauth_signature_method="HMAC-SHA1",oauth_timestamp="%s",oauth_token="%s"'''\
               ''',oauth_version="1.0"''' \
               % (email, nonce, urllib.quote(signature), timestamp, urllib.quote(a_token))




class CredentialHelper(object):
    """
       Helper handling all credentials
    """
    SECRET_FILEPATH = '%s/token.sec' 
    
    @classmethod
    def get_secret_key(cls, a_filepath):
        """
           Get secret key if it is in the file otherwise generate it and save it
        """
        if os.path.exists(a_filepath):
            secret = open(a_filepath).read()
        else:
            secret = gmvault_utils.make_password()
            
            fdesc = os.open(a_filepath, os.O_CREAT|os.O_WRONLY, 0600)
            
            the_bytes = os.write(fdesc, secret)
            os.close(fdesc) #close anyway
            
            if the_bytes < len(secret):
                raise Exception("Error: Cannot write secret in %s" % (a_filepath))

        return secret
    
    @classmethod
    def store_passwd(cls, email, passwd):
        """
           Encrypt and store gmail password
        """
        passwd_file = '%s/%s.passwd' % (gmvault_utils.get_home_dir_path(), email)
    
        fdesc = os.open(passwd_file, os.O_CREAT|os.O_WRONLY, 0600)
        
        cipher       = blowfish.Blowfish(cls.get_secret_key(cls.SECRET_FILEPATH % (gmvault_utils.get_home_dir_path())))
        cipher.initCTR()
    
        encrypted = cipher.encryptCTR(passwd)
        the_bytes = os.write(fdesc, encrypted)
    
        os.close(fdesc)
        
        if the_bytes < len(encrypted):
            raise Exception("Error: Cannot write password in %s" % (passwd_file))
        
    @classmethod
    def store_oauth_credentials(cls, email, token, secret, type):
        """
           store oauth_credentials
        """
        oauth_file = '%s/%s.oauth' % (gmvault_utils.get_home_dir_path(), email)
    
        fdesc = os.open(oauth_file, os.O_CREAT|os.O_WRONLY, 0600)
        
        os.write(fdesc, token)
        os.write(fdesc, '::')
        os.write(fdesc, secret)
        os.write(fdesc, '::')
        os.write(fdesc, type)
    
        os.close(fdesc)
    
    @classmethod
    def read_password(cls, email):
        """
           Read password credentials
           Look by default to ~/.gmvault
           Look for file ~/.gmvault/email.passwd
        """
        gmv_dir = gmvault_utils.get_home_dir_path()
        
        #look for email.passwed in GMV_DIR
        user_passwd_file_path = "%s/%s.passwd" % (gmv_dir, email)

        password = None
        if os.path.exists(user_passwd_file_path):
            passwd_file  = open(user_passwd_file_path)
            
            password     = passwd_file.read()
            cipher       = blowfish.Blowfish(cls.get_secret_key(cls.SECRET_FILEPATH % (gmvault_utils.get_home_dir_path())))
            cipher.initCTR()
            password     = cipher.decryptCTR(password)
        
        return password
    
    @classmethod
    def read_oauth_tok_sec(cls, email):
        """
           Read oauth token secret credential
           Look by default to ~/.gmvault
           Look for file ~/.gmvault/email.oauth
        """
        gmv_dir = gmvault_utils.get_home_dir_path()
        
        #look for email.passwed in GMV_DIR
        user_oauth_file_path = "%s/%s.oauth" % (gmv_dir, email)

        token  = None
        secret = None
        type   = None
        if os.path.exists(user_oauth_file_path):
            LOG.critical("Get XOAuth credential from %s.\n" % (user_oauth_file_path))
            
            oauth_file  = open(user_oauth_file_path)
            
            try:
                oauth_result = oauth_file.read()
                if oauth_result:
                    oauth_result = oauth_result.split('::')
                    if len(oauth_result) == 2:
                        token  = oauth_result[0]
                        secret = oauth_result[1]
                        type   = "normal"
                    elif len(oauth_result) == 3:
                        token  = oauth_result[0]
                        secret = oauth_result[1]
                        type   = oauth_result[2]
            except Exception, _: #pylint: disable-msg=W0703              
                LOG.critical("Cannot read oauth credentials from %s. Force oauth credentials renewal." % (user_oauth_file_path))
                LOG.critical("=== Exception traceback ===")
                LOG.critical(gmvault_utils.get_exception_traceback())
                LOG.critical("=== End of Exception traceback ===\n")
        
        if token: token   = token.strip() #pylint: disable-msg=C0321
        if secret: secret = secret.strip()  #pylint: disable-msg=C0321
        if type: type = type.strip()
        
        return token, secret, type
            
    @classmethod
    def get_credential(cls, args, test_mode = {'activate': False, 'value' : 'test_password'}): #pylint: disable-msg=W0102
        """
           Deal with the credentials.
           1) Password
           --passwd passed. If --passwd passed and not password given if no password saved go in interactive mode
           2) XOAuth Token
        """
        credential = { }
        
        #first check that there is an email
        if not args.get('email', None):
            raise Exception("No email passed, Need to pass an email")
        
        if args['passwd'] in ['empty', 'store', 'renew']: 
            # --passwd is here so look if there is a passwd in conf file 
            # or go in interactive mode
            
            LOG.critical("Authentication performed with Gmail password.\n")
            
            passwd = cls.read_password(args['email'])
            
            #password to be renewed so need an interactive phase to get the new pass
            if not passwd or args['passwd'] in ['renew', 'store']: # go to interactive mode
                if not test_mode.get('activate', False):
                    passwd = getpass.getpass('Please enter gmail password for %s and press ENTER:' % (args['email']))
                else:
                    passwd = test_mode.get('value', 'no_password_given')
                    
                credential = { 'type' : 'passwd', 'value' : passwd}
                
                #store it in dir if asked for --store-passwd or --renew-passwd
                if args['passwd'] in ['renew', 'store']:
                    LOG.critical("Store password for %s in $HOME/.gmvault." % (args['email']))
                    cls.store_passwd(args['email'], passwd)
                    credential['option'] = 'saved'
            else:
                LOG.critical("Use password stored in $HOME/.gmvault dir (Storing your password here is not recommended).")
                credential = { 'type' : 'passwd', 'value' : passwd, 'option':'read' }
                               
        #elif args['passwd'] == 'not_seen' and args['oauth']:
        elif args['passwd'] in ('not_seen', None) and args['oauth'] in (None, 'empty', 'renew', 'not_seen'):
            # get token secret
            # if they are in a file then no need to call get_oauth_tok_sec
            # will have to add 2 legged 
            LOG.critical("Authentication performed with Gmail XOAuth token.\n")
            
            two_legged = args.get('two_legged', False) # 2 legged oauth
            
            token, secret, type = cls.read_oauth_tok_sec(args['email'])
           
            if not token or args['oauth'] == 'renew':
                
                if args['oauth'] == 'renew':
                    LOG.critical("Renew XOAuth token (normal or 2-legged). Initiate interactive session to get it from Gmail.\n")
                else:
                    LOG.critical("Initiate interactive session to get XOAuth normal or 2-legged token from Gmail.\n")
                
                if two_legged:
                    token, secret, type = get_2_legged_oauth_tok_sec()
                else:
                    token, secret, type = get_oauth_tok_sec(args['email'], use_webbrowser = True)
                
                if not token:
                    raise Exception("Cannot get XOAuth token from Gmail. See Gmail error message")
                #store newly created token
                cls.store_oauth_credentials(args['email'], token, secret, type)
               
            xoauth_req = generate_xoauth_req(token, secret, args['email'], type)
            
            LOG.critical("Successfully read oauth credentials.\n")

            credential = { 'type' : 'xoauth', 'value' : xoauth_req, 'option':None }
                        
        return credential

    @classmethod
    def get_xoauth_req_from_email(cls, email):
        """
           This will be used to reconnect after a timeout
        """
        token, secret, type = cls.read_oauth_tok_sec(email)
        if not token: 
            raise Exception("Error cannot read token, secret from")
        
        xoauth_req = generate_xoauth_req(token, secret, email, type)
        
        return xoauth_req



# -*- coding: utf-8 -*-
'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
import base64
import os
import datetime
import hashlib

import gmv.gmvault as gmvault
import gmv.imap_utils       as imap_utils
import gmv.credential_utils as cred_utils
import gmv.gmvault_db as gmvault_db
import gmv.gmvault_utils    as gmvault_utils


def check_remote_mailbox_identical_to_local(the_self, gmvaulter): #pylint: disable=C0103,R0912,R0914,R0915
    """
       Check that the remote mailbox is identical to the local one attached
       to gmvaulter
       Need a connected gmvaulter
    """
    # get all email data from gmvault-db
    pivot_dir  = None
    gmail_ids  = gmvaulter.gstorer.get_all_existing_gmail_ids(pivot_dir)

    print("gmail_ids = %s\n" % (gmail_ids))
    
    #need to check that all labels are there for emails in essential
    gmvaulter.src.select_folder('ALLMAIL')
    
    # check the number of id on disk 
    imap_ids = gmvaulter.src.search({ 'type' : 'imap', 'req' : 'ALL'}) #get everything
    
    the_self.assertEquals(len(imap_ids), \
                      len(gmail_ids), \
                      "Error. Should have the same number of emails: local nb of emails %d,"\
                      " remote nb of emails %d" % (len(gmail_ids), len(imap_ids)))

    for gm_id in gmail_ids:

        print("Fetching id %s with request %s" % (gm_id, imap_utils.GIMAPFetcher.GET_ALL_BUT_DATA))
        #get disk_metadata
        disk_metadata   = gmvaulter.gstorer.unbury_metadata(gm_id)

        print("disk metadata %s\n" % (disk_metadata))

        #date     = disk_metadata['internal_date'].strftime('"%d %b %Y"')
        subject  = disk_metadata.get('subject', None)
        msgid    = disk_metadata.get('msg_id', None)
        received = disk_metadata.get('x_gmail_received', None)

        req = "("
        has_something = False

        #if date:
        #    req += 'HEADER DATE {date}'.format(date=date)
        #    has_something = True

        if subject:
            #split on ' when contained in subject to keep only the first part
            subject = subject.split("'")[0]
            subject = subject.split('"')[0]
            if has_something: #add extra space if it has a date
                req += ' ' 
            req += 'SUBJECT "{subject}"'.format(subject=subject.strip().encode('utf-8'))
            has_something = True

        if msgid:
            if has_something: #add extra space if it has a date
                req += ' ' 
            req += 'HEADER MESSAGE-ID {msgid}'.format(msgid=msgid.strip())
            has_something = True
        
        if received:
            if has_something:
                req += ' '
                req += 'HEADER X-GMAIL-RECEIVED {received}'.format(received=received.strip())
                has_something = True
        
        req += ")"

        print("Req = %s\n" % (req))

        imap_ids = gmvaulter.src.search({ 'type' : 'imap', 'req': req, 'charset': 'utf-8'})

        print("imap_ids = %s\n" % (imap_ids))

        if len(imap_ids) != 1:
            the_self.fail("more than one imap_id (%s) retrieved for request %s" % (imap_ids, req))

        imap_id = imap_ids[0]
        
        # get online_metadata 
        online_metadata = gmvaulter.src.fetch(imap_id, \
                                              imap_utils.GIMAPFetcher.GET_ALL_BUT_DATA) 

        print("online_metadata = %s\n" % (online_metadata))
        print("disk_metadata = %s\n"   % (disk_metadata))

        header_fields = online_metadata[imap_id]['BODY[HEADER.FIELDS (MESSAGE-ID SUBJECT X-GMAIL-RECEIVED)]']
        
        subject, msgid, received = gmvault_db.GmailStorer.parse_header_fields(header_fields)

        #compare metadata
        the_self.assertEquals(subject, disk_metadata.get('subject', None))
        the_self.assertEquals(msgid,   disk_metadata.get('msg_id', None))
        the_self.assertEquals(received, disk_metadata.get('x_gmail_received', None))

        # check internal date it is plus or minus 1 hour
        online_date   = online_metadata[imap_id].get('INTERNALDATE', None) 
        disk_date     = disk_metadata.get('internal_date', None) 

        if online_date != disk_date:
            min_date = disk_date - datetime.timedelta(hours=1)
            max_date = disk_date + datetime.timedelta(hours=1)
            
            if min_date <= online_date <= max_date:
                print("online_date (%s) and disk_date (%s) differs but "\
                      "within one hour. This is OK (timezone pb) *****" % (online_date, disk_date))
            else:
                the_self.fail("online_date (%s) and disk_date (%s) are different" % (online_date, disk_date))

        #check labels
        disk_labels   = disk_metadata.get('labels', None)
        online_labels = imap_utils.decode_labels(online_metadata[imap_id].get('X-GM-LABELS', None)) 

        #clean potential labels with multiple spaces
        disk_labels   = [ gmvault_utils.remove_consecutive_spaces_and_strip(label) for label in disk_labels ]
        online_labels = [ gmvault_utils.remove_consecutive_spaces_and_strip(label) for label in online_labels ]

        if not disk_labels: #no disk_labels check that there are no online_labels
            the_self.assertTrue(not online_labels)

        the_self.assertEquals(len(disk_labels), len(online_labels))

        for label in disk_labels:
            #change label Migrated (lower and uppercase) to gmv-migrated because reserved by Gmail
            if label.lower() == "migrated":
               label = "gmv-migrated"
            if label not in online_labels:
                the_self.fail("label %s should be in online_labels %s as"\
                              " it is in disk_labels %s" % (label, online_labels, disk_labels))

        # check flags
        disk_flags   = disk_metadata.get('flags', None)
        online_flags = online_metadata[imap_id].get('FLAGS', None) 

        if not disk_flags: #no disk flags
            the_self.assertTrue(not online_flags)

        the_self.assertEquals(len(disk_flags), len(online_flags))

        for flag in disk_flags:
            if flag not in online_flags:
                the_self.fail("flag %s should be in "\
                              "online_flags %s as it is in disk_flags %s" \
                              % (flag, online_flags, disk_flags))        

def find_identical_emails(gmvaulter_a): #pylint: disable=R0914
    """
       Find emails that are identical
    """
    # check all ids one by one
    gmvaulter_a.src.select_folder('ALLMAIL')
    
    # check the number of id on disk 
    imap_ids_a = gmvaulter_a.src.search({ 'type' : 'imap', 'req' : 'ALL'}) 
    
    batch_size = 1000

    batch_fetcher_a = gmvault.IMAPBatchFetcher(gmvaulter_a.src, imap_ids_a, \
                      gmvaulter_a.error_report, imap_utils.GIMAPFetcher.GET_ALL_BUT_DATA, \
                      default_batch_size = batch_size)
    
    print("Got %d emails in gmvault_a(%s).\n" % (len(imap_ids_a), gmvaulter_a.login))
    
    identicals = {}  

    in_db = {}
    
    total_processed = 0

    imap_ids = gmvaulter_a.src.search({ 'type' : 'imap', \
               'req' : '(HEADER MESSAGE-ID 1929235391.1106286872672.JavaMail.wserver@disvds016)'})

    print("Len(imap_ids): %d, imap_ids = %s" % (len(imap_ids), imap_ids))

    # get all gm_id for fetcher_b
    for gm_ids in batch_fetcher_a:
        cpt = 0
        #print("gm_ids = %s\n" % (gm_ids))
        print("Process a new batch (%d). Total processed:%d.\n" % (batch_size, total_processed))
        for one_id in gm_ids:
            if cpt % 50 == 0:
                print("look for %s" % (one_id))
            header_fields = gm_ids[one_id]['BODY[HEADER.FIELDS (MESSAGE-ID SUBJECT X-GMAIL-RECEIVED)]']
        
            subject, msgid, received = gmvault_db.GmailStorer.parse_header_fields(header_fields)
            labels        = gm_ids[one_id]['X-GM-LABELS']
            date_internal = gm_ids[one_id]['INTERNALDATE'] 

            if not in_db.get(msgid, None):
                in_db[msgid] = [{'subject': subject, 'received': received, \
                                 'gmid': gm_ids[one_id]['X-GM-MSGID'], \
                                 'date': date_internal , 'labels': labels}]  
            else:
                in_db[msgid].append({'subject': subject, 'received': received, \
                             'gmid': gm_ids[one_id]['X-GM-MSGID'], \
                             'date': date_internal , 'labels': labels}) 
                print("identical found msgid %s : %s" \
                      % (msgid, {'subject': subject, \
                        'received': received, \
                        'gmid': gm_ids[one_id]['X-GM-MSGID'],\
                        'date': date_internal , 'labels': labels}))
                
            cpt += 1 
        total_processed += batch_size

    #create list of identicals
    for msgid in in_db:
        if len(in_db[msgid]) > 1:
            identicals[msgid] = in_db[msgid]

    #print identicals
    print("Found %d identicals" % (len(identicals)))
    for msgid in identicals:
        print("== MSGID ==: %s" % (msgid))
        for vals in identicals[msgid]:
            print("===========> gmid: %s ### date: %s ### subject: %s ### "\
                  "labels: %s ### received: %s" \
                  % (vals.get('gmid',None), vals.get('date', None),\
                  vals.get('subject',None), vals.get('labels', None), \
                  vals.get('received',None)))
            #print("vals:%s" % (vals))
        print("\n")
        
    #print("Identical emails:\n%s" % (identicals))   

def diff_online_mailboxes(gmvaulter_a, gmvaulter_b): #pylint: disable=R0912, R0914
    """
       Diff 2 mailboxes
    """
    # check all ids one by one
    gmvaulter_a.src.select_folder('ALLMAIL')
    gmvaulter_b.src.select_folder('ALLMAIL')
    
    # check the number of id on disk 
    imap_ids_a = gmvaulter_a.src.search({ 'type' : 'imap', 'req' : 'ALL'}) 
    imap_ids_b = gmvaulter_b.src.search({ 'type' : 'imap', 'req' : 'ALL'}) 
    
    batch_size = 700

    batch_fetcher_a = gmvault.IMAPBatchFetcher(gmvaulter_a.src, imap_ids_a, gmvaulter_a.error_report, \
                                               imap_utils.GIMAPFetcher.GET_ALL_BUT_DATA, \
                                               default_batch_size = batch_size)
    
    batch_fetcher_b = gmvault.IMAPBatchFetcher(gmvaulter_b.src, imap_ids_b, gmvaulter_b.error_report, \
                                               imap_utils.GIMAPFetcher.GET_ALL_BUT_DATA, \
                                               default_batch_size = batch_size)
    
    print("Got %d emails in gmvault_a(%s).\n" % (len(imap_ids_a), gmvaulter_a.login))
    print("Got %d emails in gmvault_b(%s).\n" % (len(imap_ids_b), gmvaulter_b.login))
    
    if len(imap_ids_a) != len(imap_ids_b):
        print("Oh Oh, gmvault_a has %s emails and gmvault_b has %s emails\n" \
              % (len(imap_ids_a), len(imap_ids_b)))
    else:
        print("Both databases has %d emails." % (len(imap_ids_a)))
    
    diff_result = { "in_a" : {},
                    "in_b" : {},
                  }  
    
    gm_ids_b = {}
    total_processed = 0
    # get all gm_id for fetcher_b
    for gm_ids in batch_fetcher_b:
        #print("gm_ids = %s\n" % (gm_ids))
        print("Process a new batch (%d). Total processed:%d.\n" % (batch_size, total_processed))
        for one_id in gm_ids:
            gm_id = gm_ids[one_id]['X-GM-MSGID']
            
            header_fields = gm_ids[one_id]['BODY[HEADER.FIELDS (MESSAGE-ID SUBJECT X-GMAIL-RECEIVED)]']
        
            subject, msgid, received = gmvault_db.GmailStorer.parse_header_fields(header_fields)
            
            the_hash = hashlib.md5()
            if received:
                the_hash.update(received)
            
            if subject:
                the_hash.update(subject)
                
            if msgid:
                the_hash.update(msgid)

            id =  base64.encodestring(the_hash.digest())
    
            gm_ids_b[id] = [gm_id, subject, msgid]

        total_processed += batch_size

    #dumb search not optimisation
    #iterate over imap_ids_a and flag emails only in a but not in b
    #remove emails from imap_ids_b everytime they are found 
    for data_infos in batch_fetcher_a:
        for gm_info in data_infos:
            gm_id = data_infos[gm_info]['X-GM-MSGID']
            
            header_fields = data_infos[gm_info]['BODY[HEADER.FIELDS (MESSAGE-ID SUBJECT X-GMAIL-RECEIVED)]']
        
            subject, msgid, received = gmvault_db.GmailStorer.parse_header_fields(header_fields)
            
            the_hash = hashlib.md5()
            if received:
                the_hash.update(received)
            
            if subject:
                the_hash.update(subject)
                
            if msgid:
                the_hash.update(msgid)

            id =  base64.encodestring(the_hash.digest())
    
            if id not in gm_ids_b:
                diff_result["in_a"][received] = [gm_id, subject, msgid]
            else:
                del gm_ids_b[id]
    
    for recv_id in gm_ids_b:
        diff_result["in_b"][recv_id] = gm_ids_b[recv_id]
        
    
    # print report
    if (len(diff_result["in_a"]) > 0 or len(diff_result["in_b"]) > 0):
        print("emails only in gmv_a:\n") 
        print_diff_result(diff_result["in_a"])
        print("\n")
        print("emails only in gmv_b:%s\n") 
        print_diff_result(diff_result["in_b"])
    else:
        print("Mailbox %s and %s are identical.\n" % (gmvaulter_a.login, gmvaulter_b.login))
        
def print_diff_result(diff_result):
    """ print the diff_result structure
    """
    for key in diff_result:
        vals = diff_result[key]
        print("mailid:%s#####subject:%s#####%s." % (vals[2], vals[1], vals[0]))


def assert_login_is_protected(login):
    """
      Insure that the login is not my personnal mailbox
    """
    if login != 'gsync.mtester@gmail.com':
        raise Exception("Beware login should be gsync.mtester@gmail.com and it is %s" % (login)) 

def clean_mailbox(login , credential):
    """
       Delete all emails, destroy all labels
    """
    gimap = imap_utils.GIMAPFetcher('imap.gmail.com', 993, login, credential, readonly_folder = False)

    print("login = %s" % (login))

    assert_login_is_protected(login)

    gimap.connect()
    
    gimap.erase_mailbox()


def obfuscate_string(a_str):
    """ use base64 to obfuscate a string """
    return base64.b64encode(a_str)

def deobfuscate_string(a_str):
    """ deobfuscate a string """
    return base64.b64decode(a_str)

def read_password_file(a_path):
    """
       Read log:pass from a file in my home
    """
    pass_file = open(a_path)
    line = pass_file.readline()
    (login, passwd) = line.split(":")
    
    return (deobfuscate_string(login.strip()), deobfuscate_string(passwd.strip()))

def get_oauth_cred(email, cred_path):
    """
       Read oauth token secret credential
       Look by default to ~/.gmvault
       Look for file ~/.gmvault/email.oauth
    """
    user_oauth_file_path = cred_path

    token  = None
    secret = None
    if os.path.exists(user_oauth_file_path):
        print("Get XOAuth credential from %s.\n" % (user_oauth_file_path))
             
        oauth_file  = open(user_oauth_file_path)
             
        try:
            oauth_result = oauth_file.read()
            if oauth_result:
                oauth_result = oauth_result.split('::')
                if len(oauth_result) == 2:
                    token  = oauth_result[0]
                    secret = oauth_result[1]
        except Exception, _: #pylint: disable-msg=W0703              
            print("Cannot read oauth credentials from %s. Force oauth credentials renewal." % (user_oauth_file_path))
            print("=== Exception traceback ===")
            print(gmvault_utils.get_exception_traceback())
            print("=== End of Exception traceback ===\n")
         
        if token: token   = token.strip() #pylint: disable-msg=C0321
        if secret: secret = secret.strip()  #pylint: disable-msg=C0321
 
    return { 'type' : 'xoauth', 'value' : cred_utils.generate_xoauth_req(token, secret, email, 'normal'), 'option':None}

def delete_db_dir(a_db_dir):
    """
       delete the db directory
    """
    gmvault_utils.delete_all_under(a_db_dir, delete_top_dir = True)



'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
# exception classes
class Error(Exception):
    """Base class for Conf exceptions."""

    def __init__(self, msg=''):
        self.message = msg
        Exception.__init__(self, msg)

    def __repr__(self):
        return self.message

    __str__ = __repr__
    
class NoOptionError(Error):
    """A requested option was not found."""

    def __init__(self, option, section):
        Error.__init__(self, "No option %r in section: %r" % 
                       (option, section))
        self.option = option
        self.section = section

class NoSectionError(Error):
    """Raised when no section matches a requested option."""

    def __init__(self, section):
        Error.__init__(self, 'No section: %r' % (section,))
        self.section = section

class SubstitutionError(Error):
    """Base class for substitution-related exceptions."""

    def __init__(self, lineno, location, msg):
        Error.__init__(self, 'SubstitutionError on line %d: %s. %s' \
                       % (lineno, location, msg) if lineno != - 1 \
                       else 'SubstitutionError in %s. %s' % (lineno, location))
        
class IncludeError(Error):
    """ Raised when an include command is incorrect """
    
    def __init__(self, msg, origin):
        Error.__init__(self, msg)
        self.origin = origin
        self.errors = []


class ParsingError(Error):
    """Raised when a configuration file does not follow legal syntax."""
    def __init__(self, filename):
        Error.__init__(self, 'File contains parsing errors: %s' % filename)
        self.filename = filename
        self.errors = []

    def append(self, lineno, line):
        """ add error message """
        self.errors.append((lineno, line))
        self.message += '\n\t[line %2d]: %s' % (lineno, line)
        
    def get_error(self):
        """ return the error """
        return self
        
class MissingSectionHeaderError(ParsingError):
    """Raised when a key-value pair is found before any section header."""

    def __init__(self, filename, lineno, line):
        ParsingError.__init__(
            self,
            'File contains no section headers.\nfile: %s, line: %d\n%r' % 
            (filename, lineno, line))
        self.filename = filename
        self.lineno = lineno
        self.line = line



'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
import json
import time
import datetime
import os
import itertools
import imaplib

import gmv.log_utils as log_utils
import gmv.collections_utils as collections_utils
import gmv.gmvault_utils as gmvault_utils
import gmv.imap_utils as imap_utils
import gmv.gmvault_db as gmvault_db

LOG = log_utils.LoggerFactory.get_logger('gmvault')

def handle_restore_imap_error(the_exception, gm_id, db_gmail_ids_info, gmvaulter):
    """
       function to handle restore IMAPError in restore functions 
    """
    if isinstance(the_exception, imaplib.IMAP4.abort):
        # if this is a Gmvault SSL Socket error quarantine the email and continue the restore
        if str(the_exception).find("=> Gmvault ssl socket error: EOF") >= 0:
            LOG.critical("Quarantine email with gm id %s from %s. GMAIL IMAP cannot restore it:"\
                         " err={%s}" % (gm_id, db_gmail_ids_info[gm_id], str(the_exception)))
            gmvaulter.gstorer.quarantine_email(gm_id)
            gmvaulter.error_report['emails_in_quarantine'].append(gm_id)
            LOG.critical("Disconnecting and reconnecting to restart cleanly.")
            gmvaulter.src.reconnect() #reconnect
        else:
            raise the_exception
        
    elif isinstance(the_exception, imaplib.IMAP4.error): 
        LOG.error("Catched IMAP Error %s" % (str(the_exception)))
        LOG.exception(the_exception)
        
        #When the email cannot be read from Database because it was empty when returned by gmail imap
        #quarantine it.
        if str(the_exception) == "APPEND command error: BAD ['Invalid Arguments: Unable to parse message']":
            LOG.critical("Quarantine email with gm id %s from %s. GMAIL IMAP cannot restore it:"\
                         " err={%s}" % (gm_id, db_gmail_ids_info[gm_id], str(the_exception)))
            gmvaulter.gstorer.quarantine_email(gm_id)
            gmvaulter.error_report['emails_in_quarantine'].append(gm_id) 
        else:
            raise the_exception
    elif isinstance(the_exception, imap_utils.PushEmailError):
        LOG.error("Catch the following exception %s" % (str(the_exception)))
        LOG.exception(the_exception)
        
        if the_exception.quarantined():
            LOG.critical("Quarantine email with gm id %s from %s. GMAIL IMAP cannot restore it:"\
                         " err={%s}" % (gm_id, db_gmail_ids_info[gm_id], str(the_exception)))
            gmvaulter.gstorer.quarantine_email(gm_id)
            gmvaulter.error_report['emails_in_quarantine'].append(gm_id) 
        else:
            raise the_exception          
    else:
        LOG.error("Catch the following exception %s" % (str(the_exception)))
        LOG.exception(the_exception)
        raise the_exception

def handle_sync_imap_error(the_exception, the_id, error_report, src):
    """
      function to handle IMAPError in gmvault
      type = chat or email
    """    
    if isinstance(the_exception, imaplib.IMAP4.abort):
        # imap abort error 
        # ignore it 
        # will have to do something with these ignored messages
        LOG.critical("Error while fetching message with imap id %s." % (the_id))
        LOG.critical("\n=== Exception traceback ===\n")
        LOG.critical(gmvault_utils.get_exception_traceback())
        LOG.critical("=== End of Exception traceback ===\n")
        try:
            #try to get the gmail_id
            curr = src.fetch(the_id, imap_utils.GIMAPFetcher.GET_GMAIL_ID) 
        except Exception, _: #pylint:disable-msg=W0703
            curr = None
            LOG.critical("Error when trying to get gmail id for message with imap id %s." % (the_id))
            LOG.critical("Disconnect, wait for 20 sec then reconnect.")
            src.disconnect()
            #could not fetch the gm_id so disconnect and sleep
            #sleep 10 sec
            time.sleep(10)
            LOG.critical("Reconnecting ...")
            src.connect()
            
        if curr:
            gmail_id = curr[the_id][imap_utils.GIMAPFetcher.GMAIL_ID]
        else:
            gmail_id = None
            
        #add ignored id
        error_report['cannot_be_fetched'].append((the_id, gmail_id))
        
        LOG.critical("Forced to ignore message with imap id %s, (gmail id %s)." \
                     % (the_id, (gmail_id if gmail_id else "cannot be read")))
        
    elif isinstance(the_exception, imaplib.IMAP4.error):
        # check if this is a cannot be fetched error 
        # I do not like to do string guessing within an exception but I do not have any choice here
        LOG.critical("Error while fetching message with imap id %s." % (the_id))
        LOG.critical("\n=== Exception traceback ===\n")
        LOG.critical(gmvault_utils.get_exception_traceback())
        LOG.critical("=== End of Exception traceback ===\n")
         
        #quarantine emails that have raised an abort error
        if str(the_exception).find("'Some messages could not be FETCHed (Failure)'") >= 0:
            try:
                #try to get the gmail_id
                LOG.critical("One more attempt. Trying to fetch the Gmail ID for %s" % (the_id) )
                curr = src.fetch(the_id, imap_utils.GIMAPFetcher.GET_GMAIL_ID) 
            except Exception, _: #pylint:disable-msg=W0703
                curr = None
            
            if curr:
                gmail_id = curr[the_id][imap_utils.GIMAPFetcher.GMAIL_ID]
            else:
                gmail_id = None
            
            #add ignored id
            error_report['cannot_be_fetched'].append((the_id, gmail_id))
            
            LOG.critical("Ignore message with imap id %s, (gmail id %s)" % (the_id, (gmail_id if gmail_id else "cannot be read")))
        
        else:
            raise the_exception #rethrow error
    else:
        raise the_exception    

class IMAPBatchFetcher(object):
    """
       Fetch IMAP data in batch 
    """
    def __init__(self, src, imap_ids, error_report, request, default_batch_size = 100):
        """
           constructor
        """
        self.src                = src
        self.imap_ids           = imap_ids
        self.def_batch_size     = default_batch_size
        self.request            = request
        self.error_report       = error_report  
        
        self.to_fetch           = list(imap_ids)
    
    def individual_fetch(self, imap_ids):
        """
           Find the imap_id creating the issue
           return the data related to the imap_ids
        """
        new_data = {}
        for the_id in imap_ids:    
            try: 
                single_data = self.src.fetch(the_id, self.request)
                new_data.update(single_data)                
            except Exception, error:
                handle_sync_imap_error(error, the_id, self.error_report, self.src) #do everything in this handler

        return new_data
   
    def __iter__(self):
        return self     
    
    def next(self):
        """
            Return the next batch of elements
        """
        new_data = {}
        batch = self.to_fetch[:self.def_batch_size]
        
        if len(batch) <= 0:
            raise StopIteration
        
        try:
        
            new_data = self.src.fetch(batch, self.request)
            
            self.to_fetch = self.to_fetch[self.def_batch_size:]
            
            return new_data

        except imaplib.IMAP4.error, _:
            new_data = self.individual_fetch(batch) 
    
        return new_data
    
    def reset(self):
        """
           Restart from the beginning
        """
        self.to_fetch = self.imap_ids              
               
class GMVaulter(object):
    """
       Main object operating over gmail
    """ 
    NB_GRP_OF_ITEMS         = 1400
    EMAIL_RESTORE_PROGRESS  = 'email_last_id.restore'
    CHAT_RESTORE_PROGRESS   = 'chat_last_id.restore'
    EMAIL_SYNC_PROGRESS     = 'email_last_id.sync'
    CHAT_SYNC_PROGRESS      = 'chat_last_id.sync'
    
    OP_EMAIL_RESTORE = "EM_RESTORE"
    OP_EMAIL_SYNC    = "EM_SYNC"
    OP_CHAT_RESTORE  = "CH_RESTORE"
    OP_CHAT_SYNC    = "CH_SYNC"
    
    OP_TO_FILENAME = { OP_EMAIL_RESTORE : EMAIL_RESTORE_PROGRESS,
                       OP_EMAIL_SYNC    : EMAIL_SYNC_PROGRESS,
                       OP_CHAT_RESTORE  : CHAT_RESTORE_PROGRESS,
                       OP_CHAT_SYNC     : CHAT_SYNC_PROGRESS
                     }
    
    
    def __init__(self, db_root_dir, host, port, login, \
                 credential, read_only_access = True, use_encryption = False): #pylint:disable-msg=R0913,R0914
        """
           constructor
        """   
        self.db_root_dir = db_root_dir
        
        #create dir if it doesn't exist
        gmvault_utils.makedirs(self.db_root_dir)
        
        #keep track of login email
        self.login = login
            
        # create source and try to connect
        self.src = imap_utils.GIMAPFetcher(host, port, login, credential, \
                                           readonly_folder = read_only_access)
        
        self.src.connect()
        
        LOG.debug("Connected")
        
        self.use_encryption = use_encryption
        
        #to report gmail imap problems
        self.error_report = { 'empty' : [] ,
                              'cannot_be_fetched'  : [],
                              'emails_in_quarantine' : [],
                              'reconnections' : 0}
        
        #instantiate gstorer
        self.gstorer =  gmvault_db.GmailStorer(self.db_root_dir, self.use_encryption)
        
        #timer used to mesure time spent in the different values
        self.timer = gmvault_utils.Timer()
        
    @classmethod
    def get_imap_request_btw_2_dates(cls, begin_date, end_date):
        """
           Return the imap request for those 2 dates
        """
        imap_req = 'Since %s Before %s' % (gmvault_utils.datetime2imapdate(begin_date), gmvault_utils.datetime2imapdate(end_date))
        
        return imap_req
    
    def get_operation_report(self):
        """
           Return the error report
        """
        the_str = "\n================================================================\n"\
                  "%s operation performed in %s.\n" \
                  "Number of reconnections: %d.\nNumber of emails quarantined: %d.\n" \
                  "Number of emails that could not be fetched: %d.\n" \
                  "Number of emails that were returned empty by gmail: %d\n"\
                  "================================================================" \
              % (self.error_report['operation'], \
                 self.error_report['operation_time'], \
                 self.error_report['reconnections'], \
                 len(self.error_report['emails_in_quarantine']), \
                 len(self.error_report['cannot_be_fetched']), \
                 len(self.error_report['empty'])
                )
              
        LOG.debug("error_report complete structure = %s" % (self.error_report))
        
        return the_str
        
    @classmethod
    def _get_next_date(cls, a_current_date, start_month_beginning = False):
        """
           return the next date necessary to build the imap req
        """
        if start_month_beginning:
            dummy_date   = a_current_date.replace(day=1)
        else:
            dummy_date   = a_current_date
            
        # the next date = current date + 1 month
        return dummy_date + datetime.timedelta(days=31)
        
    @classmethod
    def check_email_on_disk(cls, a_gstorer, a_id, a_dir = None):
        """
           Factory method to create the object if it exists
        """
        try:
            a_dir = a_gstorer.get_directory_from_id(a_id, a_dir)
           
            if a_dir:
                return a_gstorer.unbury_metadata(a_id, a_dir) 
            
        except ValueError, json_error:
            LOG.exception("Cannot read file %s. Try to fetch the data again" % ('%s.meta' % (a_id)), json_error )
        
        return None
    
    @classmethod
    def _metadata_needs_update(cls, curr_metadata, new_metadata, chat_metadata = False):
        """
           Needs update
        """
        if curr_metadata[gmvault_db.GmailStorer.ID_K] != new_metadata['X-GM-MSGID']:
            raise Exception("Gmail id has changed for %s" % (curr_metadata['id']))
                
        #check flags   
        prev_set = set(new_metadata['FLAGS'])    
        
        for flag in curr_metadata['flags']:
            if flag not in prev_set:
                return True
            else:
                prev_set.remove(flag)
        
        if len(prev_set) > 0:
            return True
        
        #check labels
        prev_labels = set(new_metadata['X-GM-LABELS'])
        
        if chat_metadata: #add gmvault-chats labels
            prev_labels.add(gmvault_db.GmailStorer.CHAT_GM_LABEL)
            
        
        for label in curr_metadata['labels']:
            if label not in prev_labels:
                return True
            else:
                prev_labels.remove(label)
        
        if len(prev_labels) > 0:
            return True
        
        return False
    
    
    def _check_email_db_ownership(self, ownership_control):
        """
           Check email database ownership.
           If ownership control activated then fail if a new additional owner is added.
           Else if no ownership control allow one more user and save it in the list of owners
           
           Return the number of owner this will be used to activate or not the db clean.
           Activating a db cleaning on a multiownership db would be a catastrophy as it would delete all
           the emails from the others users.
        """
        #check that the gmvault-db is not associated with another user
        db_owners = self.gstorer.get_db_owners()
        if ownership_control:
            if len(db_owners) > 0 and self.login not in db_owners: #db owner should not be different unless bypass activated
                raise Exception("The email database %s is already associated with one or many logins: %s."\
                                " Use option (-m, --multiple-db-owner) if you want to link it with %s" \
                                % (self.db_root_dir, ", ".join(db_owners), self.login))
        else:
            if len(db_owners) == 0:
                LOG.critical("Establish %s as the owner of the Gmvault db %s." % (self.login, self.db_root_dir))  
            elif len(db_owners) > 0 and self.login not in db_owners:
                LOG.critical("The email database %s is hosting emails from %s. It will now also store emails from %s" \
                             % (self.db_root_dir, ", ".join(db_owners), self.login))
                
        #try to save db_owner in the list of owners
        self.gstorer.store_db_owner(self.login)
        
    def _sync_chats(self, imap_req, compress, restart):
        """
           sync emails
        """
        chat_dir = None
        
        timer = gmvault_utils.Timer() #start local timer for chat
        timer.start()
        
        LOG.debug("Before selection")
        if self.src.is_visible('CHATS'):
            chat_dir = self.src.select_folder('CHATS')
        
        LOG.debug("Selection is finished")

        if chat_dir:
            imap_ids = self._common_sync(timer, "chat", imap_req, compress, restart)
        else:
            imap_ids = []    
        
        LOG.critical("\nchats synchronisation operation performed in %s.\n" % (timer.seconds_to_human_time(timer.elapsed())))

        return imap_ids


    def _common_sync(self, a_timer, a_type, imap_req, compress, restart):
        """
           common syncing method for both emails and chats. 
        """
        # get all imap ids in All Mail
        imap_ids = self.src.search(imap_req)
        
        # check if there is a restart
        if restart:
            LOG.critical("Restart mode activated for emails. Need to find information in Gmail, be patient ...")
            imap_ids = self.get_gmails_ids_left_to_sync(self.OP_EMAIL_SYNC if a_type == "email" \
                                                        else self.OP_CHAT_SYNC, imap_ids, imap_req)
        
        total_nb_msgs_to_process = len(imap_ids) # total number of emails to get
        
        LOG.critical("%d %ss to be fetched." % (total_nb_msgs_to_process, a_type))
        
        nb_msgs_processed = 0
        
        to_fetch = set(imap_ids)
        batch_fetcher = IMAPBatchFetcher(self.src, imap_ids, self.error_report, imap_utils.GIMAPFetcher.GET_ALL_BUT_DATA, \
                                         default_batch_size = \
                                         gmvault_utils.get_conf_defaults().getint("General","nb_messages_per_batch",500))
        
        #choose different bury methods if it is an email or a chat
        if a_type == "email":
            bury_metadata_fn = self.gstorer.bury_metadata
            bury_data_fn     = self.gstorer.bury_email
            chat_metadata    = False
        elif a_type == "chat":
            bury_metadata_fn = self.gstorer.bury_chat_metadata
            bury_data_fn     = self.gstorer.bury_chat
            chat_metadata    = True
        else:
            raise Exception("Error a_type %s in _common_sync is unknown" % (a_type))
        
        #LAST Thing to do remove all found ids from imap_ids and if ids left add missing in report
        for new_data in batch_fetcher:            
            for the_id in new_data:
                if new_data.get(the_id, None):
                    LOG.debug("\nProcess imap id %s" % ( the_id ))
                        
                    gid      = new_data[the_id][imap_utils.GIMAPFetcher.GMAIL_ID]
                    eml_date = new_data[the_id][imap_utils.GIMAPFetcher.IMAP_INTERNALDATE]
                    
                    if a_type == "email":
                        the_dir = gmvault_utils.get_ym_from_datetime(eml_date)
                    elif a_type == "chat":
                        the_dir = self.gstorer.get_sub_chats_dir()
                    else:
                        raise Exception("Error a_type %s in _common_sync is unknown" % (a_type))
                    
                    LOG.critical("Process %s num %d (imap_id:%s) from %s." % (a_type, nb_msgs_processed, the_id, the_dir))
                    
                    #decode the labels that are received as utf7 => unicode
                    new_data[the_id][imap_utils.GIMAPFetcher.GMAIL_LABELS] = \
                    imap_utils.decode_labels(new_data[the_id][imap_utils.GIMAPFetcher.GMAIL_LABELS])

                    LOG.debug("metadata info collected: %s\n" % (new_data[the_id]))
                
                    #pass the dir and the ID
                    curr_metadata = GMVaulter.check_email_on_disk( self.gstorer , \
                                                                   new_data[the_id][imap_utils.GIMAPFetcher.GMAIL_ID], \
                                                                   the_dir)
                    
                    #if on disk check that the data is not different
                    if curr_metadata:
                        
                        LOG.debug("metadata for %s already exists. Check if different." % (gid))
                        
                        if self._metadata_needs_update(curr_metadata, new_data[the_id], chat_metadata):
                            
                            LOG.debug("%s with imap id %s and gmail id %s has changed. Updated it." % (a_type, the_id, gid))
                            
                            #restore everything at the moment
                            gid  = bury_metadata_fn(new_data[the_id], local_dir = the_dir)
                            
                            #update local index id gid => index per directory to be thought out
                        else:
                            LOG.debug("On disk metadata for %s is up to date." % (gid))
                    else:  
                        try:
                            #get the data
                            LOG.debug("Get Data for %s." % (gid))
                            email_data = self.src.fetch(the_id, imap_utils.GIMAPFetcher.GET_DATA_ONLY )
                            
                            new_data[the_id][imap_utils.GIMAPFetcher.EMAIL_BODY] = \
                            email_data[the_id][imap_utils.GIMAPFetcher.EMAIL_BODY]
                            
                            LOG.debug("Storing on disk data for %s" % (gid))
                            # store data on disk within year month dir 
                            gid  = bury_data_fn(new_data[the_id], local_dir = the_dir, compress = compress)
                            
                            #update local index id gid => index per directory to be thought out
                            LOG.debug("Create and store email with imap id %s, gmail id %s." % (the_id, gid))   
                        except Exception, error:
                            handle_sync_imap_error(error, the_id, self.error_report, self.src) #do everything in this handler    
                    
                    nb_msgs_processed += 1
                    
                    #indicate every 50 messages the number of messages left to process
                    left_emails = (total_nb_msgs_to_process - nb_msgs_processed)
                    
                    if (nb_msgs_processed % 50) == 0 and (left_emails > 0):
                        elapsed = a_timer.elapsed() #elapsed time in seconds
                        LOG.critical("\n== Processed %d emails in %s. %d left to be stored (time estimate %s).==\n" % \
                                     (nb_msgs_processed,  \
                                      a_timer.seconds_to_human_time(elapsed), left_emails, \
                                      a_timer.estimate_time_left(nb_msgs_processed, elapsed, left_emails)))
                    
                    # save id every 10 restored emails
                    if (nb_msgs_processed % 10) == 0:
                        if gid:
                            self.save_lastid(self.OP_EMAIL_SYNC, gid, eml_date, imap_req)
                else:
                    LOG.info("Could not process message with id %s. Ignore it\n" % (the_id))
                    self.error_report['empty'].append((the_id, gid if gid else None))
                    
            to_fetch -= set(new_data.keys()) #remove all found keys from to_fetch set
                
        for the_id in to_fetch:
            # case when gmail IMAP server returns OK without any data whatsoever
            # eg. imap uid 142221L ignore it
            LOG.info("Could not process imap with id %s. Ignore it\n" % (the_id))
            self.error_report['empty'].append((the_id, None))
        
        return imap_ids

    def _sync_emails(self, imap_req, compress, restart):
        """
           sync emails
        """
        timer = gmvault_utils.Timer()
        timer.start()

        #select all mail folder using the constant name defined in GIMAPFetcher
        self.src.select_folder('ALLMAIL')

        imap_ids = self._common_sync(timer, "email", imap_req, compress, restart)

        LOG.critical("\nEmails synchronisation operation performed in %s.\n" % (timer.seconds_to_human_time(timer.elapsed())))

        return imap_ids

        

    def sync(self, imap_req, compress_on_disk = True, \
             db_cleaning = False, ownership_checking = True, \
            restart = False, emails_only = False, chats_only = False):
        """
           sync mode 
        """
        #check ownership to have one email per db unless user wants different
        #save the owner if new
        self._check_email_db_ownership(ownership_checking)
          
        if not compress_on_disk:
            LOG.critical("Disable compression when storing emails.")
            
        if self.use_encryption:
            LOG.critical("Encryption activated. All emails will be encrypted before to be stored.")
            LOG.critical("Please take care of the encryption key stored in (%s) or all"\
                         " your stored emails will become unreadable." \
                         % (gmvault_db.GmailStorer.get_encryption_key_path(self.db_root_dir)))
        
        self.error_report['operation'] = 'Sync'
        
        self.timer.start() #start syncing emails
        
        now = datetime.datetime.now()
        LOG.critical("Start synchronization (%s).\n" % (now.strftime('%Y-%m-%dT%Hh%Mm%Ss')))
        
        if not chats_only:
            # backup emails
            LOG.critical("Start emails synchronization.")
            self._sync_emails(imap_req, compress = compress_on_disk, restart = restart)
        else:
            LOG.critical("Skip emails synchronization.\n")
        
        if not emails_only:
            # backup chats
            LOG.critical("Start chats synchronization.")
            self._sync_chats(imap_req, compress = compress_on_disk, restart = restart)
        else:
            LOG.critical("\nSkip chats synchronization.\n")
        
        #delete supress emails from DB since last sync
        self.check_clean_db(db_cleaning)
       
        LOG.debug("Sync operation performed in %s.\n" \
                     % (self.timer.seconds_to_human_time(self.timer.elapsed())))
        self.error_report["operation_time"] = self.timer.seconds_to_human_time(self.timer.elapsed())
        
        #update number of reconnections
        self.error_report["reconnections"] = self.src.total_nb_reconns
        
        return self.error_report

    
    def _delete_sync(self, imap_ids, db_gmail_ids, db_gmail_ids_info, msg_type):
        """
           Delete emails or chats from the database if necessary
           imap_ids      : all remote imap_ids to check
           db_gmail_ids_info : info read from metadata
           msg_type : email or chat
        """
        
        # optimize nb of items
        nb_items = self.NB_GRP_OF_ITEMS if len(imap_ids) >= self.NB_GRP_OF_ITEMS else len(imap_ids)
        
        LOG.critical("Call Gmail to check the stored %ss against the Gmail %ss ids and see which ones have been deleted.\n\n"\
                     "This might take a few minutes ...\n" % (msg_type, msg_type)) 
         
        #calculate the list elements to delete
        #query nb_items items in one query to minimise number of imap queries
        for group_imap_id in itertools.izip_longest(fillvalue=None, *[iter(imap_ids)]*nb_items):
            
            # if None in list remove it
            if None in group_imap_id: 
                group_imap_id = [ im_id for im_id in group_imap_id if im_id != None ]
            
            data = self.src.fetch(group_imap_id, imap_utils.GIMAPFetcher.GET_GMAIL_ID)
            
            # syntax for 2.7 set comprehension { data[key][imap_utils.GIMAPFetcher.GMAIL_ID] for key in data }
            # need to create a list for 2.6
            db_gmail_ids.difference_update([data[key][imap_utils.GIMAPFetcher.GMAIL_ID] for key in data ])
            
            if len(db_gmail_ids) == 0:
                break
        
        LOG.critical("Will delete %s %s(s) from gmvault db.\n" % (len(db_gmail_ids), msg_type) )
        for gm_id in db_gmail_ids:
            LOG.critical("gm_id %s not in the Gmail server. Delete it." % (gm_id))
            self.gstorer.delete_emails([(gm_id, db_gmail_ids_info[gm_id])], msg_type)
        
    def search_on_date(self, a_eml_date):
        """
           get eml_date and format it to search 
        """
        imap_date = gmvault_utils.datetime2imapdate(a_eml_date)
        
        imap_req = "SINCE %s" % (imap_date)

        imap_ids = self.src.search({'type':'imap', 'req': imap_req})
        
        return imap_ids
            
    def get_gmails_ids_left_to_sync(self, op_type, imap_ids, imap_req):#pylint:disable-msg=W0613
        """
           Get the ids that still needs to be sync
           Return a list of ids
        """
        filename = self.OP_TO_FILENAME.get(op_type, None)
        
        if not filename:
            raise Exception("Bad Operation (%s) in save_last_id. "\
                  "This should not happen, send the error to the software developers." % (op_type))
        
        filepath = '%s/%s_%s' % (self.gstorer.get_info_dir(), self.login, filename)
        
        if not os.path.exists(filepath):
            LOG.critical("last_id.sync file %s doesn't exist.\nSync the full list of backed up emails." %(filepath))
            return imap_ids
        
        json_obj = json.load(open(filepath, 'r'))
        
        last_id = json_obj['last_id']
        
        last_id_index = -1
        
        new_gmail_ids = imap_ids
        
        try:
            #get imap_id from stored gmail_id
            dummy = self.src.search({'type':'imap', 'req':'X-GM-MSGID %s' % (last_id)})
            
            imap_id = dummy[0]
            
            last_id_index = imap_ids.index(imap_id)
            
            LOG.critical("Restart from gmail id %s (imap id %s)." % (last_id, imap_id))
            
            new_gmail_ids = imap_ids[last_id_index:]   
        except Exception, _: #ignore any exception and try to get all ids in case of problems. pylint:disable=W0703
            #element not in keys return current set of keys
            LOG.critical("Error: Cannot restore from last restore gmail id. It is not in Gmail."\
                         " Sync the complete list of gmail ids requested from Gmail.")
        
        return new_gmail_ids
        
    def check_clean_db(self, db_cleaning):
        """
           Check and clean the database (remove file that are not anymore in Gmail)
        """
        owners = self.gstorer.get_db_owners()
        if not db_cleaning: #decouple the 2 conditions for activating cleaning
            LOG.debug("db_cleaning is off so ignore removing deleted emails from disk.")
            return
        elif len(owners) > 1:
            LOG.critical("The Gmvault db hosts emails from the following accounts: %s.\n"\
                         % (", ".join(owners)))
            
            LOG.critical("Deactivate database cleaning on a multi-owners Gmvault db.")
        
            return
        else:
            LOG.critical("Look for emails/chats that are in the Gmvault db but not in Gmail servers anymore.\n")
            
            #get gmail_ids from db
            LOG.critical("Read all gmail ids from the Gmvault db. It might take a bit of time ...\n")
            
            timer = gmvault_utils.Timer() # needed for enhancing the user information
            timer.start()
            
            db_gmail_ids_info = self.gstorer.get_all_existing_gmail_ids()
        
            LOG.critical("Found %s email(s) in the Gmvault db.\n" % (len(db_gmail_ids_info)) )
        
            #create a set of keys
            db_gmail_ids = set(db_gmail_ids_info.keys())
            
            # get all imap ids in All Mail
            self.src.select_folder('ALLMAIL') #go to all mail
            imap_ids = self.src.search(imap_utils.GIMAPFetcher.IMAP_ALL) #search all
            
            LOG.debug("Got %s emails imap_id(s) from the Gmail Server." % (len(imap_ids)))
            
            #delete supress emails from DB since last sync
            self._delete_sync(imap_ids, db_gmail_ids, db_gmail_ids_info, 'email')
            
            # get all chats ids
            if self.src.is_visible('CHATS'):
            
                db_gmail_ids_info = self.gstorer.get_all_chats_gmail_ids()
                
                LOG.critical("Found %s chat(s) in the Gmvault db.\n" % (len(db_gmail_ids_info)) )
                
                self.src.select_folder('CHATS') #go to chats
                chat_ids = self.src.search(imap_utils.GIMAPFetcher.IMAP_ALL)
                
                db_chat_ids = set(db_gmail_ids_info.keys())
                
                LOG.debug("Got %s chat imap_ids from the Gmail Server." % (len(chat_ids)))
            
                #delete supress emails from DB since last sync
                self._delete_sync(chat_ids, db_chat_ids, db_gmail_ids_info , 'chat')
            else:
                LOG.critical("Chats IMAP Directory not visible on Gmail. Ignore deletion of chats.")
                
            
            LOG.critical("\nDeletion checkup done in %s." % (timer.elapsed_human_time()))
            
    
    def remote_sync(self):
        """
           Sync with a remote source (IMAP mirror or cloud storage area)
        """
        #sync remotely 
        pass
        
    
    def save_lastid(self, op_type, gm_id, eml_date = None, imap_req = None):#pylint:disable-msg=W0613
        """
           Save the passed gmid in last_id.restore
           For the moment reopen the file every time
        """
        
        filename = self.OP_TO_FILENAME.get(op_type, None)
        
        if not filename:
            raise Exception("Bad Operation (%s) in save_last_id. "\
                            "This should not happen, send the error to the software developers." % (op_type))
        
        filepath = '%s/%s_%s' % (self.gstorer.get_info_dir(), self.login, filename)  
        
        the_fd = open(filepath, 'w')
        
        #json.dump({
        #            'last_id' : gm_id,
        #            'date'    : gmvault_utils.datetime2e(eml_date) if eml_date else None,
        #            'req'     : imap_req 
        #          }, the_fd)
        
        json.dump({
                    'last_id' : gm_id,
                  }, the_fd)
        
        the_fd.close()
        
    def get_gmails_ids_left_to_restore(self, op_type, db_gmail_ids_info):
        """
           Get the ids that still needs to be restored
           Return a dict key = gm_id, val = directory
        """
        filename = self.OP_TO_FILENAME.get(op_type, None)
        
        if not filename:
            raise Exception("Bad Operation (%s) in save_last_id. This should not happen,"\
                            " send the error to the software developers." % (op_type))
        
        
        #filepath = '%s/%s_%s' % (gmvault_utils.get_home_dir_path(), self.login, filename)
        filepath = '%s/%s_%s' % (self.gstorer.get_info_dir(), self.login, filename)
        
        if not os.path.exists(filepath):
            LOG.critical("last_id restore file %s doesn't exist.\nRestore the full list of backed up emails." %(filepath))
            return db_gmail_ids_info
        
        json_obj = json.load(open(filepath, 'r'))
        
        last_id = json_obj['last_id']
        
        last_id_index = -1
        try:
            keys = db_gmail_ids_info.keys()
            last_id_index = keys.index(last_id)
            LOG.critical("Restart from gmail id %s." % (last_id))
        except ValueError, _:
            #element not in keys return current set of keys
            LOG.error("Cannot restore from last restore gmail id. It is not in the disk database.")
        
        new_gmail_ids_info = collections_utils.OrderedDict()
        if last_id_index != -1:
            for key in db_gmail_ids_info.keys()[last_id_index+1:]:
                new_gmail_ids_info[key] =  db_gmail_ids_info[key]
        else:
            new_gmail_ids_info = db_gmail_ids_info    
            
        return new_gmail_ids_info 
           
    def restore(self, pivot_dir = None, extra_labels = [], \
                restart = False, emails_only = False, chats_only = False): #pylint:disable=W0102
        """
           Restore emails in a gmail account
        """
        
        self.error_report['operation'] = 'Sync'
        self.timer.start() #start restoring
        
        now = datetime.datetime.now()
        LOG.critical("Start restoration (%s).\n" % (now.strftime('%Y-%m-%dT%Hh%Mm%Ss')))
        
        if not chats_only:
            # backup emails
            LOG.critical("Start emails restoration.\n")
            
            if pivot_dir:
                LOG.critical("Quick mode activated. Will only restore all emails since %s.\n" % (pivot_dir))
            
            self.restore_emails(pivot_dir, extra_labels, restart)
        else:
            LOG.critical("Skip emails restoration.\n")
        
        if not emails_only:
            # backup chats
            LOG.critical("Start chats restoration.\n")
            self.restore_chats(extra_labels, restart)
        else:
            LOG.critical("Skip chats restoration.\n")
        
        LOG.debug("Restore operation performed in %s.\n" \
                     % (self.timer.seconds_to_human_time(self.timer.elapsed())))
        
        self.error_report["operation_time"] = self.timer.seconds_to_human_time(self.timer.elapsed())
        
        #update number of reconnections
        self.error_report["reconnections"] = self.src.total_nb_reconns
        
        return self.error_report
       
    def restore_chats(self, extra_labels = [], restart = False): #pylint:disable=W0102
        """
           restore chats
        """
        LOG.critical("Restore chats in gmail account %s." % (self.login) ) 
                
        LOG.critical("Read chats info from %s gmvault-db." % (self.db_root_dir))
        
        #get gmail_ids from db
        db_gmail_ids_info = self.gstorer.get_all_chats_gmail_ids()
        
        LOG.critical("Total number of chats to restore %s." % (len(db_gmail_ids_info.keys())))
        
        if restart:
            db_gmail_ids_info = self.get_gmails_ids_left_to_restore(self.OP_CHAT_RESTORE, db_gmail_ids_info)
        
        total_nb_emails_to_restore = len(db_gmail_ids_info)
        LOG.critical("Got all chats id left to restore. Still %s chats to do.\n" % (total_nb_emails_to_restore) )
        
        existing_labels     = set() #set of existing labels to not call create_gmail_labels all the time
        reserved_labels     = gmvault_utils.get_conf_defaults().get_list("Restore", "reserved_labels", [ u'migrated' ])
        nb_emails_restored  = 0  #to count nb of emails restored
        labels_to_apply     = collections_utils.SetMultimap()

        #get all mail folder name
        all_mail_name = self.src.get_folder_name("ALLMAIL")
        
        # go to DRAFTS folder because if you are in ALL MAIL when uploading emails it is very slow
        folder_def_location = gmvault_utils.get_conf_defaults().get("General", "restore_default_location", "DRAFTS")
        self.src.select_folder(folder_def_location)
        
        timer = gmvault_utils.Timer() # local timer for restore emails
        timer.start()
        
        nb_items = gmvault_utils.get_conf_defaults().get_int("General", "nb_messages_per_restore_batch", 100) 
        
        for group_imap_ids in itertools.izip_longest(fillvalue=None, *[iter(db_gmail_ids_info)]*nb_items): 

            last_id = group_imap_ids[-1] #will be used to save the last id
            #remove all None elements from group_imap_ids
            group_imap_ids = itertools.ifilter(lambda x: x != None, group_imap_ids)
           
            labels_to_create    = set() #create label set
            labels_to_create.update(extra_labels) # add extra labels to applied to all emails
            
            LOG.critical("Processing next batch of %s chats.\n" % (nb_items))
            
            # unbury the metadata for all these emails
            for gm_id in group_imap_ids:    
                email_meta, email_data = self.gstorer.unbury_email(gm_id)
                
                LOG.critical("Pushing chat content with id %s." % (gm_id))
                LOG.debug("Subject = %s." % (email_meta[self.gstorer.SUBJECT_K]))
                try:
                    # push data in gmail account and get uids
                    imap_id = self.src.push_data(all_mail_name, email_data, \
                                    email_meta[self.gstorer.FLAGS_K] , \
                                    email_meta[self.gstorer.INT_DATE_K] )      
                
                    #labels for this email => real_labels U extra_labels
                    labels = set(email_meta[self.gstorer.LABELS_K])
                    
                    # add in the labels_to_create struct
                    for label in labels:
                        LOG.debug("label = %s\n" % (label))
                        labels_to_apply[str(label)] = imap_id
            
                    # get list of labels to create (do a union with labels to create)
                    labels_to_create.update([ label for label in labels if label not in existing_labels])                  
                
                except Exception, err:
                    handle_restore_imap_error(err, gm_id, db_gmail_ids_info, self)

            #create the non existing labels and update existing labels
            if len(labels_to_create) > 0:
                LOG.debug("Labels creation tentative for chats ids %s." % (group_imap_ids))
                existing_labels = self.src.create_gmail_labels(labels_to_create, existing_labels)
                
            # associate labels with emails
            LOG.critical("Applying labels to the current batch of chats.")
            try:
                LOG.debug("Changing directory. Going into ALLMAIL")
                self.src.select_folder('ALLMAIL') #go to ALL MAIL to make STORE usable
                for label in labels_to_apply.keys():
                    if label.lower() in reserved_labels: #exclude creation of migrated label
                        LOG.info("Apply label 'gmv-migrated' instead of '%s' (lower or uppercase)"\
                                 " because it is a Gmail reserved label." % (label)) 
                        #need ot change labels_to_apply
                        ids = labels_to_apply[label]
                        del labels_to_apply[label]
                        label = "gmv-migrated"
                        labels_to_apply[label] = ids
                    self.src.apply_labels_to(labels_to_apply[label], [label]) 
            except Exception, err:
                LOG.error("Problem when applying labels %s to the following ids: %s" %(label, labels_to_apply[label]), err)
                if isinstance(err, imaplib.IMAP4.abort) and str(err).find("=> Gmvault ssl socket error: EOF") >= 0:
                    # if this is a Gmvault SSL Socket error ignore labelling and continue the restore
                    LOG.critical("Ignore labelling")
                    LOG.critical("Disconnecting and reconnecting to restart cleanly.")
                    self.src.reconnect() #reconnect
                else:
                    raise err
            finally:
                self.src.select_folder(folder_def_location) # go back to an empty DIR (Drafts) to be fast
                labels_to_apply = collections_utils.SetMultimap() #reset label to apply
            
            nb_emails_restored += nb_items
                
            #indicate every 10 messages the number of messages left to process
            left_emails = (total_nb_emails_to_restore - nb_emails_restored)
            
            if (left_emails > 0): 
                elapsed = timer.elapsed() #elapsed time in seconds
                LOG.critical("\n== Processed %d chats in %s. %d left to be restored "\
                             "(time estimate %s).==\n" % \
                             (nb_emails_restored, timer.seconds_to_human_time(elapsed), \
                              left_emails, timer.estimate_time_left(nb_emails_restored, elapsed, left_emails)))
            
            # save id every nb_items restored emails
            # add the last treated gm_id
            self.save_lastid(self.OP_EMAIL_RESTORE, last_id)
            
        return self.error_report 
                    
    def restore_emails(self, pivot_dir = None, extra_labels = [], restart = False):
        """
           restore emails in a gmail account using batching to group restore
           If you are not in "All Mail" Folder, it is extremely fast to push emails.
           But it is not possible to reapply labels if you are not in All Mail because the uid which is returned
           is dependant on the folder. On the other hand, you can restore labels in batch which would help gaining lots of time.
           The idea is to get a batch of 50 emails and push them all in the mailbox one by one and get the uid for each of them.
           Then create a dict of labels => uid_list and for each label send a unique store command after having changed dir
        """
        LOG.critical("Restore emails in gmail account %s." % (self.login) ) 
        
        LOG.critical("Read email info from %s gmvault-db." % (self.db_root_dir))
        
        #get gmail_ids from db
        db_gmail_ids_info = self.gstorer.get_all_existing_gmail_ids(pivot_dir)
        
        LOG.critical("Total number of elements to restore %s." % (len(db_gmail_ids_info.keys())))
        
        if restart:
            db_gmail_ids_info = self.get_gmails_ids_left_to_restore(self.OP_EMAIL_RESTORE, db_gmail_ids_info)
        
        total_nb_emails_to_restore = len(db_gmail_ids_info)
        
        LOG.critical("Got all emails id left to restore. Still %s emails to do.\n" % (total_nb_emails_to_restore) )
        
        existing_labels     = set() #set of existing labels to not call create_gmail_labels all the time
        reserved_labels     = gmvault_utils.get_conf_defaults().get_list("Restore", "reserved_labels", [ u'migrated' ])
        nb_emails_restored  = 0  #to count nb of emails restored
        labels_to_apply     = collections_utils.SetMultimap()

        #get all mail folder name
        all_mail_name = self.src.get_folder_name("ALLMAIL")
        
        # go to DRAFTS folder because if you are in ALL MAIL when uploading emails it is very slow
        folder_def_location = gmvault_utils.get_conf_defaults().get("General", "restore_default_location", "DRAFTS")
        self.src.select_folder(folder_def_location)
        
        timer = gmvault_utils.Timer() # local timer for restore emails
        timer.start()
        
        nb_items = gmvault_utils.get_conf_defaults().get_int("General", "nb_messages_per_restore_batch", 80) 
        
        for group_imap_ids in itertools.izip_longest(fillvalue=None, *[iter(db_gmail_ids_info)]*nb_items): 
            
            last_id = group_imap_ids[-1] #will be used to save the last id
            #remove all None elements from group_imap_ids
            group_imap_ids = itertools.ifilter(lambda x: x != None, group_imap_ids)
           
            labels_to_create    = set() #create label set
            labels_to_create.update(extra_labels) # add extra labels to applied to all emails
            
            LOG.critical("Processing next batch of %s emails.\n" % (nb_items))
            
            # unbury the metadata for all these emails
            for gm_id in group_imap_ids:    
                email_meta, email_data = self.gstorer.unbury_email(gm_id)
                
                LOG.critical("Pushing email body with id %s." % (gm_id))
                LOG.debug("Subject = %s." % (email_meta[self.gstorer.SUBJECT_K]))
                try:
                    # push data in gmail account and get uids
                    imap_id = self.src.push_data(all_mail_name, email_data, \
                                    email_meta[self.gstorer.FLAGS_K] , \
                                    email_meta[self.gstorer.INT_DATE_K] )      
                
                    #labels for this email => real_labels U extra_labels
                    labels = set(email_meta[self.gstorer.LABELS_K])
                    
                    # add in the labels_to_create struct
                    for label in labels:
                        if label != "\\Starred":
                            #LOG.debug("label = %s\n" % (label.encode('utf-8')))
                            LOG.debug("label = %s\n" % (label))
                            if label.lower() in reserved_labels: #exclude creation of migrated label
                                LOG.debug("Apply label 'gmv-migrated' instead of '%s' (lower or uppercase)"\
                                          " because it is a Gmail reserved label."\
                                          % (label) ) 
                                label = "gmv-migrated"
                            labels_to_apply[label] = imap_id
            
                    # get list of labels to create (do a union with labels to create)
                    labels_to_create.update([ label for label in labels if label not in existing_labels])                  
                
                except Exception, err:
                    handle_restore_imap_error(err, gm_id, db_gmail_ids_info, self)

            #create the non existing labels and update existing labels
            if len(labels_to_create) > 0:
                LOG.debug("Labels creation tentative for emails with ids %s." % (group_imap_ids))
                existing_labels = self.src.create_gmail_labels(labels_to_create, existing_labels)
                
            # associate labels with emails
            LOG.critical("Applying labels to the current batch of emails.")
            try:
                LOG.debug("Changing directory. Going into ALLMAIL")
                the_timer = gmvault_utils.Timer()
                the_timer.start()
                self.src.select_folder('ALLMAIL') #go to ALL MAIL to make STORE usable
                LOG.debug("Changed dir. Operation time = %s ms" % (the_timer.elapsed_ms()))
                for label in labels_to_apply.keys():
                    self.src.apply_labels_to(labels_to_apply[label], [label]) 
            except Exception, err:
                LOG.error("Problem when applying labels %s to the following ids: %s" %(label, labels_to_apply[label]), err)
                LOG.error("Problem when applying labels.", err)
                if isinstance(err, imaplib.IMAP4.abort) and str(err).find("=> Gmvault ssl socket error: EOF") >= 0:
                    # if this is a Gmvault SSL Socket error ignore labelling and continue the restore
                    LOG.critical("Ignore labelling")
                    LOG.critical("Disconnecting and reconnecting to restart cleanly.")
                    self.src.reconnect() #reconnect
                else:
                    raise err
            finally:
                self.src.select_folder(folder_def_location) # go back to an empty DIR (Drafts) to be fast
                labels_to_apply = collections_utils.SetMultimap() #reset label to apply
            
            nb_emails_restored += nb_items
                
            #indicate every 10 messages the number of messages left to process
            left_emails = (total_nb_emails_to_restore - nb_emails_restored)
            
            if (left_emails > 0): 
                elapsed = timer.elapsed() #elapsed time in seconds
                LOG.critical("\n== Processed %d emails in %s. %d left to be restored "\
                             "(time estimate %s). ==\n" % \
                             (nb_emails_restored, timer.seconds_to_human_time(elapsed), \
                              left_emails, timer.estimate_time_left(nb_emails_restored, elapsed, left_emails)))
            
            # save id every 50 restored emails
            # add the last treated gm_id
            self.save_lastid(self.OP_EMAIL_RESTORE, last_id)
            
        return self.error_report 
        



'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2012>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
'''
   Export function of Gmvault created by dave@vasilevsky.ca
'''

import os
import re
import mailbox

import imapclient.imap_utf7 as imap_utf7

import gmv.imap_utils as imap_utils
import gmv.log_utils as log_utils
import gmv.gmvault_utils as gmvault_utils
import gmv.gmvault_db as gmvault_db

LOG = log_utils.LoggerFactory.get_logger('gmvault_export')

class GMVaultExporter(object):
    """
       Class hanlding the creation of exports in standard formats
       such as maildir, mbox
    """
    PROGRESS_INTERVAL = 200

    CHATS_FOLDER = 'Chats'
    ARCHIVED_FOLDER = 'Archived' # Mails only in 'All Mail'

    GM_ALL = re.sub(r'^\\', '', imap_utils.GIMAPFetcher.GENERIC_GMAIL_ALL)
    GM_INBOX = 'Inbox'
    GM_SEP = '/'

    GM_SEEN = '\\Seen'
    GM_FLAGGED = '\\Flagged'

    def __init__(self, db_dir, mailbox, labels = None):
        """
           constructor
        """
        self.storer = gmvault_db.GmailStorer(db_dir)
        self.mailbox = mailbox
        self.labels = labels

    def want_label(self, label):
        """ helper indicating is a label is needed"""
        if self.labels:
            return label in self.labels
        return label != self.GM_ALL

    def export(self):
        """core method for starting the export """
        self.export_ids('emails', self.storer.get_all_existing_gmail_ids(), \
            default_folder = self.GM_ALL, use_labels = True)
        self.export_ids('chats', self.storer.get_all_chats_gmail_ids(), \
            default_folder = self.CHATS_FOLDER, use_labels = False)

    def printable_label_list(self, labels):
        """helper to print a list of labels"""
        labels = [l.encode('ascii', 'backslashreplace') for l in labels]
        return u'; '.join(labels)

    def export_ids(self, kind, ids, default_folder, use_labels):
        """ export organised by ids """
        exported_labels = "default labels"
        if self.labels:
            exported_labels = "labels " + self.printable_label_list(self.labels)
        LOG.critical("Start %s export for %s." % (kind, exported_labels))

        timer = gmvault_utils.Timer()
        timer.start()
        done = 0

        for a_id in ids:
            meta, msg = self.storer.unbury_email(a_id)

            folders = [default_folder]
            if use_labels:
                add_labels = meta[gmvault_db.GmailStorer.LABELS_K]
                if not add_labels:
                    add_labels = [GMVaultExporter.ARCHIVED_FOLDER]
                folders.extend(add_labels)
            folders = [re.sub(r'^\\', '', f) for f in folders]
            folders = [f for f in folders if self.want_label(f)]

            LOG.debug("Processing id %s in labels %s." % \
                (a_id, self.printable_label_list(folders)))
            for folder in folders:
                self.mailbox.add(msg, folder, meta[gmvault_db.GmailStorer.FLAGS_K])

            done += 1
            left = len(ids) - done
            if done % self.PROGRESS_INTERVAL == 0 and left > 0:
                elapsed = timer.elapsed()
                LOG.critical("== Processed %d %s in %s, %d left (time estimate %s). ==\n" % \
                    (done, kind, timer.seconds_to_human_time(elapsed), \
                     left, timer.estimate_time_left(done, elapsed, left)))

        LOG.critical("Export completed in %s." % (timer.elapsed_human_time(),))


class Mailbox(object):
    """ Mailbox abstract class"""
    def add(self, msg, folder, flags):
        raise NotImplementedError('implement in subclass')
    def close(self):
        pass

class Maildir(Mailbox):
    """ Class delaing with the Maildir format """
    def __init__(self, path, separator = '/'):
        self.path = path
        self.subdirs = {}
        self.separator = separator
        if not self.root_is_maildir() and not os.path.exists(self.path):
            os.makedirs(self.path)

    @staticmethod
    def separate(folder, sep):
        """ separate method """
        return folder.replace(GMVaultExporter.GM_SEP, sep)

    def subdir_name(self, folder):
        """get subdir_name """
        return self.separate(folder, self.separator)

    def root_is_maildir(self):
        """check if root is maildir"""
        return False;

    def subdir(self, folder):
        """ return a subdir """
        if folder in self.subdirs:
            return self.subdirs[folder]

        if folder:
            parts = folder.split(GMVaultExporter.GM_SEP)
            parent = GMVaultExporter.GM_SEP.join(parts[:-1])
            self.subdir(parent)
            path = self.subdir_name(folder)
            path = imap_utf7.encode(path)
        else:
            if not self.root_is_maildir():
                return
            path = ''

        abspath = os.path.join(self.path, path)
        sub = mailbox.Maildir(abspath, create = True)
        self.subdirs[folder] = sub
        return sub

    def add(self, msg, folder, flags):
        """ add message in a given subdir """
        mmsg = mailbox.MaildirMessage(msg)

        if GMVaultExporter.GM_SEEN in flags:
            mmsg.set_subdir('cur')
            mmsg.add_flag('S')
        if mmsg.get_subdir() == 'cur' and GMVaultExporter.GM_FLAGGED in flags:
            mmsg.add_flag('F')

        self.subdir(folder).add(mmsg)

class OfflineIMAP(Maildir):
    """ Class dealing with offlineIMAP specificities """
    DEFAULT_SEPARATOR = '.'
    def __init__(self, path, separator = DEFAULT_SEPARATOR):
        super(OfflineIMAP, self).__init__(path, separator = separator)

class Dovecot(Maildir):
    """ Class dealing with Dovecot specificities """
    # See http://wiki2.dovecot.org/Namespaces
    class Layout(object):
        def join(self, parts):
            return self.SEPARATOR.join(parts)
    class FSLayout(Layout):
        SEPARATOR = '/'
    class MaildirPlusPlusLayout(Layout):
        SEPARATOR = '.'
        def join(self, parts):
            return '.' + super(Dovecot.MaildirPlusPlusLayout, self).join(parts)

    DEFAULT_NS_SEP = '.'
    DEFAULT_LISTESCAPE = '\\'

    # The namespace separator cannot be escaped with listescape.
    # Replace it with a two-character escape code.
    DEFAULT_SEP_ESCAPE = "*'"

    def __init__(self, path,
                 layout = MaildirPlusPlusLayout(),
                 ns_sep = DEFAULT_NS_SEP,
                 listescape = DEFAULT_LISTESCAPE,
                 sep_escape = DEFAULT_SEP_ESCAPE):
        super(Dovecot, self).__init__(path, separator = layout.SEPARATOR)
        self.layout = layout
        self.ns_sep = ns_sep
        self.listescape = listescape
        self.sep_escape = sep_escape

    # Escape one character
    def _listescape(self, s, char = None, pattern = None):
        pattern = pattern or re.escape(char)
        esc = "%s%02x" % (self.listescape, ord(char))
        return re.sub(pattern, lambda m: esc, s)

    def _munge_name(self, s):
        # Escape namespace separator: . => *', * => **
        esc = self.sep_escape[0]
        s = re.sub(re.escape(esc), esc * 2, s)
        s = re.sub(re.escape(self.ns_sep), self.sep_escape, s)

        if self.listescape:
            # See http://wiki2.dovecot.org/Plugins/Listescape
            if self.layout.SEPARATOR == '.':
                s = self._listescape(s, '.')
            s = self._listescape(s, '/')
            s = self._listescape(s, '~', r'^~')
        return s

    def subdir_name(self, folder):
        if folder == GMVaultExporter.GM_INBOX:
            return ''

        parts = folder.split(GMVaultExporter.GM_SEP)
        parts = [self._munge_name(n) for n in parts]
        return self.layout.join(parts)

    def root_is_maildir(self):
        return True

class MBox(Mailbox):
    """ Class dealing with MBox specificities """
    def __init__(self, folder):
        self.folder = folder
        self.open = dict()

    def close(self):
        for _, m in self.open.items():
            m.close()

    def subdir(self, label):
        segments = label.split(GMVaultExporter.GM_SEP)
        # Safety first: No unusable directory portions
        segments = [s for s in segments if s != '..' and s != '.']
        real_label = GMVaultExporter.GM_SEP.join(segments)
        if real_label in self.open:
            return self.open[real_label]

        cur_path = self.folder
        label_segments = []
        for s in segments:
            label_segments.append(s)
            cur_label = GMVaultExporter.GM_SEP.join(label_segments)
            if cur_label not in self.open:
                # Create an mbox for intermediate folders, to satisfy
                # Thunderbird import
                if not os.path.exists(cur_path):
                    os.makedirs(cur_path)
                mbox_path = os.path.join(cur_path, s)
                self.open[cur_label] = mailbox.mbox(mbox_path)
            # Use .sbd folders a la Thunderbird, to allow nested folders
            cur_path = os.path.join(cur_path, s + '.sbd')

        return self.open[real_label]

    def add(self, msg, folder, flags):
        mmsg = mailbox.mboxMessage(msg)
        if GMVaultExporter.GM_SEEN in flags:
            mmsg.add_flag('R')
        if GMVaultExporter.GM_FLAGGED in flags:
            mmsg.add_flag('F')
        self.subdir(folder).add(mmsg)



# -*- coding: utf-8 -*-
'''
    Gmvault: a tool to backup and restore your gmail account.
    Copyright (C) <2011-2013>  <guillaume Aubert (guillaume dot aubert at gmail do com)>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
import os

import re
import datetime
import time
import calendar
import fnmatch
import functools

import StringIO
import sys
import traceback
import random 
import locale

import gmv.log_utils as log_utils
import gmv.conf.conf_helper
import gmv.gmvault_const as gmvault_const

LOG = log_utils.LoggerFactory.get_logger('gmvault_utils')

GMVAULT_VERSION = "1.8-beta"

class memoized(object): #pylint: disable=C0103
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.
    """
    def __init__(self, func):
        self.func = func
        self.cache = {}
    def __call__(self, *args):
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            self.cache[args] = value
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args)
    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__
    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)
    
class Curry:
    """ Class used to implement the currification (functional programming technic) :
        Create a function from another one by instanciating some of its parameters.
        For example double = curry(operator.mul,2), res = double(4) = 8
    """
    def __init__(self, fun, *args, **kwargs):
        self.fun = fun
        self.pending = args[:]
        self.kwargs = kwargs.copy()
        
    def __call__(self, *args, **kwargs):
        if kwargs and self.kwargs:
            the_kw = self.kwargs.copy()
            the_kw.update(kwargs)
        else:
            the_kw = kwargs or self.kwargs
        return self.fun(*(self.pending + args), **the_kw) #pylint: disable=W0142



LETTERS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
DIGITS = '0123456789'

def make_password(minlength=8, maxlength=16):  
    """
       generate randomw password
    """
    length = random.randint(minlength, maxlength)
    letters = LETTERS + DIGITS 
    return ''.join([random.choice(letters) for _ in range(length)])  



def get_exception_traceback():
    """
            return the exception traceback (stack info and so on) in a string
        
            Args:
               None
               
            Returns:
               return a string that contains the exception traceback
        
            Raises:
               
    """
   
    the_file = StringIO.StringIO()
    exception_type, exception_value, exception_traceback = sys.exc_info() #IGNORE:W0702
    traceback.print_exception(exception_type, exception_value, exception_traceback, file = the_file)
    return the_file.getvalue()


MULTI_SPACES_PATTERN = r"\s{2,}"
MULTI_SPACES_RE = re.compile(MULTI_SPACES_PATTERN, flags=re.U) #to support unicode

def remove_consecutive_spaces_and_strip(a_str):
    """
       Supress consecutive spaces to replace them with a unique one.
       e.g "two  spaces" = "two spaces"
    """
    #return re.sub("\s{2,}", " ", a_str, flags=re.U).strip()
    return MULTI_SPACES_RE.sub(u" ", a_str).strip()

def buffered_write(fd, data, buf_size = 1048576):
    """
       fd: file descriptor where to write
       data: data to write
       buf_size: size of the buffer (default 1MB)
       buffered write handling case when write returns nb of written bytes
       and when write returns None (Linux)
    """
    #LOG.critical("======= In buffered write")
    total_size = len(data)
    wr_bytes = 0
    while wr_bytes < total_size:
        written = fd.write(buffer(data, wr_bytes, buf_size))
        if written:
            wr_bytes += written
        else:
            #if buffer size > left then left else buffer size
            # else buffer size
            left = total_size - wr_bytes
            if left < buf_size:
                wr_bytes += left
            else:
                wr_bytes += buf_size
        
             

TIMER_SUFFIXES = ['y', 'w', 'd', 'h', 'm', 's']

class Timer(object):
    """
       Timer Class to mesure time.
       Possess also few time utilities
    """
    
 
    def __init__(self):
        
        self._start = None
        
    def start(self):
        """
           start the timer
        """
        self._start = time.time()
        
    def reset(self):
        """
           reset the timer to 0
        """
        self._start = time.time()
    
    def elapsed(self):
        """
           return elapsed time in sec
        """
        now = time.time()
        
        return int(round(now - self._start))
    
    def elapsed_ms(self):
        """
          return elapsed time up to micro second
        """
        return time.time() - self._start
    
    def elapsed_human_time(self, suffixes=TIMER_SUFFIXES, add_s=False, separator=' '):#pylint:disable=W0102
        """
        Takes an amount of seconds and turns it into a human-readable amount of time.
        """
        seconds = self.elapsed()
        
        return self.seconds_to_human_time(seconds, suffixes, add_s, separator)

    @classmethod
    def estimate_time_left(cls, nb_elem_done, in_sec, still_to_be_done, in_human_time = True):
        """
           Stupid estimate. Use current time to estimate how long it will take
        """
        if in_human_time:
            return cls.seconds_to_human_time(int(round(float(still_to_be_done * in_sec)/nb_elem_done)))
        else:
            return int(round(float(still_to_be_done * in_sec)/nb_elem_done))
    
    @classmethod
    def seconds_to_human_time(cls, seconds, suffixes=TIMER_SUFFIXES, add_s=False, separator=' '):#pylint:disable=W0102
        """
           convert seconds to human time
        """
        # the formatted time string to be returned
        the_time = []
        
        # the pieces of time to iterate over (days, hours, minutes, etc)
        # - the first piece in each tuple is the suffix (d, h, w)
        # - the second piece is the length in seconds (a day is 60s * 60m * 24h)
        parts = [(suffixes[0], 60 * 60 * 24 * 7 * 52),
              (suffixes[1], 60 * 60 * 24 * 7),
              (suffixes[2], 60 * 60 * 24),
              (suffixes[3], 60 * 60),
              (suffixes[4], 60),
              (suffixes[5], 1)]
        
        if seconds < 1: #less than a second case
            return "less than a second"
        
        # for each time piece, grab the value and remaining seconds, and add it to
        # the time string
        for suffix, length in parts:
            value = seconds / length
            if value > 0:
                seconds = seconds % length
                the_time.append('%s%s' % (str(value),
                               (suffix, (suffix, suffix + 's')[value > 1])[add_s]))
            if seconds < 1:
                break
        
        return separator.join(the_time)

ZERO = datetime.timedelta(0) 
# A UTC class.    
class UTC(datetime.tzinfo):    
    """UTC Timezone"""    
    
    def utcoffset(self, a_dt): #pylint: disable=W0613
        ''' return utcoffset '''  
        return ZERO    
    
    def tzname(self, a_dt): #pylint: disable=W0613
        ''' return tzname '''    
        return "UTC"    
        
    def dst(self, a_dt): #pylint: disable=W0613 
        ''' return dst '''      
        return ZERO  

# pylint: enable-msg=W0613    
UTC_TZ = UTC()

def get_ym_from_datetime(a_datetime):
    """
       return year month from datetime
    """
    if a_datetime:
        return a_datetime.strftime('%Y-%m')
    
    return None

MONTH_CONV = { 1: 'Jan', 4: 'Apr', 6: 'Jun', 7: 'Jul', 10: 'Oct' , 12: 'Dec',
               2: 'Feb', 5: 'May', 8: 'Aug', 9: 'Sep', 11: 'Nov',
               3: 'Mar'}

REVERSE_MONTH_CONV = { 'Jan' : 1, 'Apr' : 4, 'Jun' : 6, 'Jul': 7, 'Oct': 10 , 'Dec':12,
                   'Feb' : 2, 'May' : 5, 'Aug' : 8, 'Sep': 9, 'Nov': 11,
                   'Mar' : 3}


MONTH_YEAR_PATTERN = r'(?P<year>(18|19|[2-5][0-9])\d\d)[-/.](?P<month>(0[1-9]|1[012]|[1-9]))'
MONTH_YEAR_RE = re.compile(MONTH_YEAR_PATTERN)

def compare_yymm_dir(first, second):
    """
       Compare directory names in the form of Year-Month
       Return 1 if first > second
              0 if equal
              -1 if second > first
    """
    
    matched = MONTH_YEAR_RE.match(first)
    
    if matched:
        first_year  = int(matched.group('year'))
        first_month = int(matched.group('month'))
        
        first_val   = (first_year * 1000) + first_month
    else:
        raise Exception("Invalid Year-Month expression (%s). Please correct it to be yyyy-mm" % (first))
        
    matched = MONTH_YEAR_RE.match(second)
    
    if matched:
        second_year  = int(matched.group('year'))
        second_month = int(matched.group('month'))
        
        second_val   = (second_year * 1000) + second_month
    else:
        raise Exception("Invalid Year-Month expression (%s). Please correct it" % (second))
    
    if first_val > second_val:
        return 1
    elif first_val == second_val:
        return 0
    else:
        return -1
    
def cmp_to_key(mycmp):
    """
        Taken from functools. Not in all python versions so had to redefine it
        Convert a cmp= function into a key= function
    """
    class Key(object): #pylint: disable=R0903
        """Key class"""
        def __init__(self, obj, *args): #pylint: disable=W0613
            self.obj = obj
        def __lt__(self, other):
            return mycmp(self.obj, other.obj) < 0
        def __gt__(self, other):
            return mycmp(self.obj, other.obj) > 0
        def __eq__(self, other):
            return mycmp(self.obj, other.obj) == 0
        def __le__(self, other):
            return mycmp(self.obj, other.obj) <= 0
        def __ge__(self, other):
            return mycmp(self.obj, other.obj) >= 0
        def __ne__(self, other):
            return mycmp(self.obj, other.obj) != 0
        def __hash__(self):
            raise TypeError('hash not implemented')
    return Key
    
def get_all_dirs_posterior_to(a_dir, dirs):
    """
           get all directories posterior
    """
    #sort the passed dirs list and return all dirs posterior to a_dir
         
    return [ name for name in sorted(dirs, key=cmp_to_key(compare_yymm_dir))\
             if compare_yymm_dir(a_dir, name) <= 0 ]

def get_all_dirs_under(root_dir, ignored_dirs = []):#pylint:disable=W0102
    """
       Get all directory names under (1 level only) the root dir
       params:
          root_dir   : the dir to look under
          ignored_dir: ignore the dir if it is in this list of dirnames 
    """
    return [ name for name in os.listdir(root_dir) \
             if ( os.path.isdir(os.path.join(root_dir, name)) \
                and (name not in ignored_dirs) ) ]

def datetime2imapdate(a_datetime):
    """
       Transfrom in date format for IMAP Request
    """
    if a_datetime:
        
        month = MONTH_CONV[a_datetime.month]
        
        pattern = '%%d-%s-%%Y' % (month) 
        
        return a_datetime.strftime(pattern)
    

def e2datetime(a_epoch):
    """
        convert epoch time in datetime

            Args:
               a_epoch: the epoch time to convert

            Returns: a datetime
    """

    #utcfromtimestamp is not working properly with a decimals.
    # use floor to create the datetime
#    decim = decimal.Decimal('%s' % (a_epoch)).quantize(decimal.Decimal('.001'), rounding=decimal.ROUND_DOWN)

    new_date = datetime.datetime.utcfromtimestamp(a_epoch)

    return new_date

def datetime2e(a_date):
    """
        convert datetime in epoch
        Beware the datetime as to be in UTC otherwise you might have some surprises
            Args:
               a_date: the datertime to convert

            Returns: a epoch time
    """
    return calendar.timegm(a_date.timetuple())

def contains_any(string, char_set):
    """Check whether 'string' contains ANY of the chars in 'set'"""
    return 1 in [c in string for c in char_set]

def makedirs(a_path):
    """ my own version of makedir """
    
    if os.path.isdir(a_path):
        # it already exists so return
        return
    elif os.path.isfile(a_path):
        raise OSError("a file with the same name as the desired dir, '%s', already exists."%(a_path))

    os.makedirs(a_path)

def __rmgeneric(path, __func__):
    """ private function that is part of delete_all_under """
    try:
        __func__(path)
        #print 'Removed ', path
    except OSError, (_, strerror): #IGNORE:W0612
        print """Error removing %(path)s, %(error)s """ % {'path' : path, 'error': strerror }
            
def delete_all_under(path, delete_top_dir = False):
    """ delete all files and directories under path """

    if not os.path.isdir(path):
        return
    
    files = os.listdir(path)

    for the_f in files:
        fullpath = os.path.join(path, the_f)
        if os.path.isfile(fullpath):
            new_f = os.remove
            __rmgeneric(fullpath, new_f)
        elif os.path.isdir(fullpath):
            delete_all_under(fullpath)
            new_f = os.rmdir
            __rmgeneric(fullpath, new_f)
    
    if delete_top_dir:
        os.rmdir(path)
        
def ordered_dirwalk(a_dir, a_file_wildcards= '*', a_dir_ignore_list = [], sort_func = sorted):#pylint:disable=W0102
    """
        Walk a directory tree, using a generator.
        This implementation returns only the files in all the subdirectories.
        Beware, this is a generator.
        Args:
        a_dir: A root directory from where to list
        a_wildcards: Filtering wildcards a la unix
    """

    
    sub_dirs = []
    for the_file in sort_func(os.listdir(a_dir)):
        fullpath = os.path.join(a_dir, the_file)
        if os.path.isdir(fullpath):
            sub_dirs.append(fullpath) #it is a sub_dir
        elif fnmatch.fnmatch(fullpath, a_file_wildcards):
            yield fullpath
        
    #iterate over sub_dirs
    for sub_dir in sort_func(sub_dirs):
        if os.path.basename(sub_dir) not in a_dir_ignore_list:
            for p_elem in ordered_dirwalk(sub_dir, a_file_wildcards):
                yield p_elem 
        else:
            LOG.debug("Ignore subdir %s" % (sub_dir))
  
def dirwalk(a_dir, a_wildcards= '*'):
    """
       return all files and dirs in a directory
    """
    for root, _, files in os.walk(a_dir):
        for the_file in files:
            if fnmatch.fnmatch(the_file, a_wildcards):
                yield os.path.join(root, the_file)  

def ascii_hex(a_str):
    """
       transform any string in hexa values
    """
    new_str = ""
    for the_char in a_str:
        new_str += "%s=hex[%s]," % (the_char, hex(ord(the_char)))
    return new_str

def profile_this(fn):
    """ profiling decorator """
    def profiled_fn(*args, **kwargs):
        import cProfile
        fpath = fn.__name__ + ".profile"
        prof  = cProfile.Profile()
        ret   = prof.runcall(fn, *args, **kwargs)
        prof.dump_stats(fpath)
        return ret
    return profiled_fn
                
def convert_to_unicode(a_str):
    """
       Try to get the stdin encoding and use it to convert the input string into unicode.
       It is dependent on the platform (mac osx,linux, windows 
    """
    #encoding can be forced from conf
    term_encoding = get_conf_defaults().get('Localisation', 'term_encoding', None)
    if not term_encoding:
        term_encoding = locale.getpreferredencoding() #use it to find the encoding for text terminal
        if not term_encoding:
            loc = locale.getdefaultlocale() #try to get defaultlocale()
            if loc and len(loc) == 2:
                term_encoding = loc[1]
            else:
                LOG.debug("Odd. loc = %s. Do not specify the encoding, let Python do its own investigation" % (loc))
    else:
        LOG.debug("Encoding forced. Read it from [Localisation]:term_encoding=%s" % (term_encoding))
        
    try: #encode
        u_str = unicode(a_str, term_encoding, errors='ignore')
           
        LOG.debug("raw unicode     = %s." % (u_str))
        LOG.debug("chosen encoding = %s." % (term_encoding))
        LOG.debug("unicode_escape val = %s." % ( u_str.encode('unicode_escape')))
    except Exception, err:
        LOG.error(err)
        get_exception_traceback()
        LOG.debug("Cannot convert to unicode from encoding:%s" % (term_encoding)) #add error
        u_str = unicode(a_str, errors='ignore')

    LOG.debug("hexval %s" % (ascii_hex(u_str)))
    
    return u_str
                
@memoized
def get_home_dir_path():
    """
       Get the gmvault dir
    """
    gmvault_dir = os.getenv("GMVAULT_DIR", None)
    
    # check by default in user[HOME]
    if not gmvault_dir:
        LOG.debug("no ENV variable $GMVAULT_DIR defined. Set by default $GMVAULT_DIR to $HOME/.gmvault (%s/.gmvault)" \
                  % (os.getenv("HOME",".")))
        gmvault_dir = "%s/.gmvault" % (os.getenv("HOME", "."))
    
    #create dir if not there
    makedirs(gmvault_dir)
    
    return gmvault_dir

CONF_FILE = "gmvault_defaults.conf"

@memoized
def get_conf_defaults():
    """
       Return the conf object containing the defaults stored in HOME/gmvault_defaults.conf
       Beware it is memoized
    """
    filepath = get_conf_filepath()
    
    if filepath:
        
        os.environ[gmv.conf.conf_helper.Conf.ENVNAME] = filepath
    
        the_cf = gmv.conf.conf_helper.Conf.get_instance()
    
        LOG.debug("Load defaults from %s" % (filepath))
        
        return the_cf
    else:
        return gmv.conf.conf_helper.MockConf() #retrun MockObject that will play defaults
    
#VERSION DETECTION PATTERN
VERSION_PATTERN  = r'\s*conf_version=\s*(?P<version>\S*)\s*'
VERSION_RE  = re.compile(VERSION_PATTERN)

#list of version conf to not overwrite with the next
VERSIONS_TO_PRESERVE = [ '1.8-beta' ]

def _get_version_from_conf(home_conf_file):
    """
       Check if the config file need to be replaced because it comes from an older version
    """
    #check version
    ver = None
    with open(home_conf_file) as curr_fd:
        for line in curr_fd:
            line = line.strip()
            matched = VERSION_RE.match(line)
            if matched:
                ver = matched.group('version')
                return ver.strip()
    
    return ver

def _create_default_conf_file(home_conf_file):
    """
       Write on disk the default file
    """
    LOG.critical("Create defaults in %s. Please touch this file only if you know what to do." % (home_conf_file))
    try:
        the_fd = open(home_conf_file, "w+")
        the_fd.write(gmvault_const.DEFAULT_CONF_FILE)
        the_fd.close()
        return home_conf_file
    except Exception, err:
        #catch all error and let run gmvault with defaults if needed
        LOG.critical("Ignore Error when trying to create conf file for defaults in %s:\n%s.\n" % (get_home_dir_path(), err) )
        LOG.debug("=== Exception traceback ===")
        LOG.debug(get_exception_traceback())
        LOG.debug("=== End of Exception traceback ===\n")
        #return default file instead
        return         

@memoized
def get_conf_filepath():
    """
       If default file is not present, generate it from scratch.
       If it cannot be created, then return None
    """
    home_conf_file = "%s/%s" % (get_home_dir_path(), CONF_FILE)
    
    if not os.path.exists(home_conf_file):
        return _create_default_conf_file(home_conf_file)
    else:
        # check if the conf file needs to be replaced
        version = _get_version_from_conf(home_conf_file)
        if version not in VERSIONS_TO_PRESERVE:
            LOG.debug("%s with version %s is too old, overwrite it with the latest file." \
                       % (home_conf_file, version))
            return _create_default_conf_file(home_conf_file)    
    
    return home_conf_file
